from collections import Counter

import MinkowskiEngine as ME
import MinkowskiFunctional as MF
import numpy as np
import torch
import torch.nn as nn

from spine.data import TensorBatch
from spine.utils.globals import (
    COORD_COLS,
    GHOST_SHP,
    PART_COL,
    PPN_LENDP_COL,
    PPN_LPART_COL,
    PPN_LTYPE_COL,
    PPN_ROFF_COLS,
    PPN_RPOS_COLS,
    PPN_RTYPE_COLS,
    SHAPE_COL,
    TRACK_SHP,
    VALUE_COL,
)
from spine.utils.logger import logger
from spine.utils.torch.scripts import cdist_fast
from spine.utils.weighting import get_class_weights

from .act_norm import act_factory
from .blocks import ASPP, SPP, ResNetBlock
from .configuration import setup_cnn_configuration

__all__ = ["PPN", "PPNLoss"]


class PPN(torch.nn.Module):
    """Point Proposal Network (PPN).

    It requires a UResNet network as a backbone. Typical configuration:

    .. code-block:: yaml

        model:
          name: uresnet_ppn_chain
          modules:
            uresnet:
              # Your uresnet config here
            ppn:
              # Your ppn config here

    Configuration
    -------------
    dimension: int, default 3
    num_input: int, default 1
    allow_bias: bool, default False
    spatial_size: int, default 512
    leakiness: float, default 0.33
    activation: dict
        For activation function, defaults to `{'name': 'lrelu', 'args': {}}`
    norm_layer: dict
        For normalization function, defaults to `{'name': 'batch_norm', 'args': {}}`

    depth: int, default 5
        Depth of UResNet, also corresponds to how many times we down/upsample.
    filters: int, default 16
        Number of filters in the first convolution of UResNet.
        Will increase linearly with depth.
    reps: int, default 2
        Convolution block repetition factor
    input_kernel: int, default 3
        Receptive field size for very first convolution after input layer.
    num_classes: int, default 5
    mask_score_threshold: float, default 0.5
    classify_endpoints: bool, default False
        Enable classification of points into start vs end points.
    ppn_resolution: float, default 1.0
    ghost: bool, default False
    use_true_ghost_mask: bool, default False
    mask_loss_name: str, default 'BCE'
        Can be 'BCE' or 'LogDice'

    Output
    ------
    ppn_points: torch.Tensor
        Contains  X, Y, Z predictions, semantic class prediction logits, and prob score
    ppn_masks: list of torch.Tensor
        Binary masks at various spatial scales of PPN predictions (voxel-wise score > some threshold)
    ppn_coords: list of torch.Tensor
        List of XYZ coordinates at various spatial scales.
    ppn_layers: list of torch.Tensor
        List of score features at various spatial scales.
    ppn_output_coords: torch.Tensor
        XYZ coordinates tensor at the very last layer of PPN (initial spatial scale)
    ppn_classify_endpoints: torch.Tensor
        Logits for end/start point classification.

    See Also
    --------
    :class:`PPNLoss`, :class:`spine.model.uresnet_ppn_chain`
    """

    def __init__(self, uresnet, ppn, uresnet_loss=None, ppn_loss=None):
        """Initializes the standalone PPN network.

        Parameters
        ----------
        uresnet : dict
            Dictionary of the backbone uresnet configuration
        ppn : dict
            Model configuration
        uresnet_loss : dict, optional
            UResNet loss configuration
        ppn_loss : dict, optional
            PPN loss configuration
        """
        # Initialize the parent class
        super().__init__()

        # Initialize the backbone configuration
        self.process_backbone_config(**uresnet)

        # Process the PPN specific parameters
        self.process_model_config(**ppn)

        # Inialize the PPN model
        self.initialize_model()

    def process_backbone_config(self, num_classes, ghost=False, **backbone):
        """Initialize the underlying UResNet model configuration.

        Parameters
        ----------
        num_classes : int
            Number of classes to classify the voxels as
        ghost : bool, default False
            Whether to add a deghosting step in the classification model
        **backbone : dict
            UResNet backbone configuration
        """
        # Store the semantic segmentation configuration
        self.num_classes = num_classes
        self.ghost = ghost

        # Initialize the UResNet backbone, store the relevant parameters
        setup_cnn_configuration(self, **backbone)

    def process_model_config(
        self,
        mask_score_threshold=0.5,
        classify_endpoints=False,
        propagate_all=False,
        use_binary_mask=False,
        ghost=False,
        use_true_ghost_mask=False,
    ):
        """Process the PPN-specific parameters.

        Parameters
        ----------
        mask_score_threshold : float, default 0.5
            Predicted score above which a pixel is considered positive
        classify_endpoints : bool, default False
            Whether or not to predict which point is the start/end for a track
        propagate_all : bool, default False
            If `True`, the mask will not be applied at every PPN layer
        use_binary_mask : bool, default False
            If `True`, converts the features to a binary mask based on score
        ghost : bool, default False
            Whether or not the input contains ghosts
        use_true_ghost_mask : bool, default False
            If `True`, ghost labels to deghost the tensor
        """
        # Register parameters
        self.mask_score_threshold = mask_score_threshold
        self.classify_endpoints = classify_endpoints
        self.propagate_all = propagate_all
        self.use_binary_mask = use_binary_mask
        self.ghost = ghost
        self.use_true_ghost_mask = use_true_ghost_mask

    def initialize_model(self):
        """Initializes the PPN-specific decoder."""
        # Initialize the decoding blocks
        self.decoding_block = []
        self.decoding_conv = []
        self.ppn_masks = nn.ModuleList()
        for i in range(self.depth - 2, -1, -1):
            m = []
            m.append(ME.MinkowskiBatchNorm(self.num_planes[i + 1]))
            m.append(act_factory(self.act_cfg))
            m.append(
                ME.MinkowskiConvolutionTranspose(
                    in_channels=self.num_planes[i + 1],
                    out_channels=self.num_planes[i],
                    kernel_size=2,
                    stride=2,
                    dimension=self.dim,
                )
            )
            m = nn.Sequential(*m)
            self.decoding_conv.append(m)

            m = []
            for j in range(self.reps):
                m.append(
                    ResNetBlock(
                        self.num_planes[i] * (2 if j == 0 else 1),
                        self.num_planes[i],
                        dimension=self.dim,
                        activation=self.act_cfg,
                    )
                )
            m = nn.Sequential(*m)
            self.decoding_block.append(m)
            self.ppn_masks.append(ME.MinkowskiLinear(self.num_planes[i], 2))

        self.decoding_block = nn.Sequential(*self.decoding_block)
        self.decoding_conv = nn.Sequential(*self.decoding_conv)

        # Expands the scores to the appropriate feature shape
        self.expand_as = ExpandAs()

        # Final ResNet block at the original image size
        num_output = self.num_planes[0]
        self.final_block = ResNetBlock(
            num_output, num_output, dimension=self.dim, activation=self.act_cfg
        )

        # Final linear layer for positional regression (dimension size)
        self.ppn_pixel_pos = ME.MinkowskiLinear(num_output, self.dim)

        # Final convolution layer for type classification
        self.ppn_type = ME.MinkowskiLinear(num_output, self.num_classes)

        # Final convolution layer for endpoint prediction
        if self.classify_endpoints:
            self.ppn_endpoint = ME.MinkowskiLinear(num_output, 2)

        # Ghost point removal tools
        if self.ghost:
            logger.debug("Ghost Masking is enabled for MinkPPN.")
            self.masker = AttentionMask()
            self.merge_concat = MergeConcat()
            self.ghost_mask = MinkGhostMask(self.dim)

    def forward(self, final_tensor, decoder_tensors, ghost=None, seg_label=None):
        """Compute the PPN loss for a batch of data.

        The PPN loss comprises three components:
        - Regression loss: position of the point of interest within a pixel
        - Type: type of point of interest
        - Mask: whether or not a pixel is within some distance of a point

        Parameters
        ----------
        final_tensor : TensorBatch
            Feature tensors at the deepest layer of the backbone UResNet
        decoder_tensors : List[TensorBatch]
            Feature tensors of each of the decoding blocks
        ghost : TensorBatch, optional
            Logits of the ghost predictions of the backbone UResNet
        seg_label : TensorBatch, optional
            Segmentation label tensor

        Returns
        -------
        dict
             Dictionary of outputs
        """
        # Get the list of decoder feature maps
        decoder_feature_maps = []
        if self.ghost:
            # If there are ghosts, must downsample the ghost label/prediction
            # and apply it to each decoder feature map
            with torch.no_grad():
                if self.use_true_ghost_mask:
                    # If using the true ghost mask, use the label tensor
                    assert seg_label is not None, (
                        "If `use_true_ghost_mask` is set to `True`, must "
                        "provide the `seg_label` tensor."
                    )

                    labels = seg_label.tensor
                    assert labels.shape[0] == decoder_tensors[-1].tensor.shape[0], (
                        "The label tensor length must match that "
                        "of the last UResNet layer"
                    )

                    ghost_coords = labels[:, :VALUE_COL]
                    ghost_mask_tensor = labels[:, SHAPE_COL] < GHOST_SHP
                else:
                    # If using predictions, convert the ghost scores to a mask
                    ghost_coords = ghost.tensor.C
                    ghost_mask_tensor = 1.0 - torch.argmax(
                        ghost.tensor.F, dim=1, keepdim=True
                    )

                ghost_mask = ME.SparseTensor(
                    ghost_mask_tensor, coordinates=ghost_coords
                )

            # Downsample stride 1 ghost mask to all intermediate decoder layers
            for t in decoder_tensors[::-1]:
                scaled_ghost_mask = self.ghost_mask(ghost_mask, t.tensor)
                nonghost_tensor = self.masker(t.tensor, scaled_ghost_mask)
                decoder_feature_maps.append(nonghost_tensor)

            decoder_feature_maps = decoder_feature_maps[::-1]

        else:
            decoder_feature_maps = [t.tensor for t in decoder_tensors]

        # Loop over the PPN decoding path
        ppn_masks, ppn_layers, ppn_coords = [], [], []
        x = final_tensor.tensor
        for i, layer in enumerate(self.decoding_conv):
            # Pass the previous features through the decoding convolution
            x = layer(x)

            # Merge with the UesNet decoding features
            decoder_tensor = decoder_feature_maps[i]
            if self.ghost:
                x = self.merge_concat(decoder_tensor, x)
            else:
                x = ME.cat(decoder_tensor, x)

            # Apply the decoding block, linear layer and sigmoid function
            x = self.decoding_block[i](x)
            scores = self.ppn_masks[i](x)
            softmax = MF.softmax(scores, dim=1)
            mask = softmax.F[:, 1:] > self.mask_score_threshold

            # Store the coordinates, raw score logits and score mask
            counts = decoder_tensors[i].counts
            ppn_coords.append(
                TensorBatch(scores.C, counts, has_batch_col=True, coord_cols=COORD_COLS)
            )
            ppn_layers.append(TensorBatch(scores.F, counts))
            ppn_masks.append(TensorBatch(mask, counts))

            # Expand the score mask
            s_expanded = self.expand_as(
                softmax,
                x.F.shape,
                propagate_all=self.propagate_all,
                use_binary_mask=self.use_binary_mask,
            )

            # Scale the out of this layer using the score mask
            x = x * s_expanded.detach()

        # Output set of coordinates (should match the last decoder tensor)
        assert x.C.shape[0] == decoder_tensors[-1].tensor.shape[0], (
            "The output of the last PPN layer should be consistent "
            "with the length of the last UResNet decoder layer"
        )
        final_counts = decoder_tensors[-1].counts
        ppn_output_coords = TensorBatch(
            x.C, final_counts, has_batch_col=True, coord_cols=COORD_COLS
        )

        # Pass the final PPN tensor through the individual predictions, combine
        x = self.final_block(x)
        pixel_pos = self.ppn_pixel_pos(x)
        ppn_type = self.ppn_type(x)
        if self.classify_endpoints:
            ppn_endpoint = self.ppn_endpoint(x)

        # X, Y, Z, logits, and prob score
        ppn_points = TensorBatch(
            torch.cat([pixel_pos.F, ppn_type.F, ppn_layers[-1].tensor], dim=1),
            final_counts,
        )

        result = {
            "ppn_points": ppn_points,
            "ppn_masks": ppn_masks,
            "ppn_layers": ppn_layers,
            "ppn_coords": ppn_coords,
            "ppn_output_coords": ppn_output_coords,
        }
        if self.classify_endpoints:
            result["ppn_classify_endpoints"] = TensorBatch(ppn_endpoint.F, final_counts)

        return result


class PPNLoss(torch.nn.modules.loss._Loss):
    """Loss function for PPN.

    Output
    ------
    reg_loss : float
        Distance loss
    mask_loss : float
        Binary voxel-wise prediction loss (is there an object of interest or not)
    classify_endpoints_loss : float
        Endpoint classification loss
    type_loss : float
        Semantic prediction loss
    output_mask_accuracy: float
        Binary voxel-wise prediction accuracy in the last layer
    type_accuracy : float
        Semantic prediction accuracy
    classify_endpoints_accuracy : float
        Endpoint classification accuracy

    See Also
    --------
    PPN, spine.model.uresnet_ppn_chain
    """

    def __init__(self, uresnet, ppn, ppn_loss, uresnet_loss=None):
        """Initializes the standalone PPN loss.

        Parameters
        ----------
        uresnet : dict
            Dictionary of the backbone uresnet configuration
        ppn : dict
            Model configuration
        uresnet_loss : dict, optional
            UResNet loss configuration
        ppn_loss : dict, optional
            PPN loss configuration
        """
        # Initialize the parent class
        super().__init__()

        # Initialize what we need from the backbone configuration
        self.process_backbone_config(**uresnet)

        # Initialize the loss configuration
        self.process_loss_config(**ppn_loss)

    def process_backbone_config(self, depth, **kwargs):
        """Process the parameters of the backbone model needed for in the loss.

        Parameters
        ----------
        depth : int
            Depth of the UResNet
        **kwargs : dict, optional
            Leftover model configuration (no need in the loss)
        """
        # Store the semantic segmentation configuration
        self.depth = depth

    def process_loss_config(
        self,
        mask_loss="CE",
        resolution=5.0,
        point_classes=None,
        balance_mask_loss=True,
        mask_weighting_mode="const",
        balance_type_loss=True,
        type_weighting_mode="const",
        reg_loss_weight=1.0,
        type_loss_weight=1.0,
        mask_loss_weight=1.0,
        endpoint_loss_weight=1.0,
        return_mask_labels=False,
        restrict_to_clusters=False,
    ):
        """Process the loss function parameters.

        Parameters
        ----------
        resolution : float, default 5.
            Distance from a label point in pixels within which a voxel is
            considered positive (pixel of interest)
        mask_loss : str, default 'CE'
            Name of the loss function to use
        point_classes : Union[int, list], optional
            If provided, restricts the loss to points of (a) certain shape(s)
        balance_mask_loss : bool, default True
            Apply class-weights to the mask loss
        mask_weighting_mode : str, default 'const'
            Method for class-weighting the mask loss
        balance_type_loss : bool, default True
            Apply class-weights to the type loss
        type_weighting_method : str, default 'const'
            Method class-weighting the type loss
        reg_loss_weight : float, default 1.
            Relative weight to apply to the regression loss
        type_loss_weight : float, default 1.
            Relative weight to apply to the point type loss
        mask_loss_weight : float, default 1.
            Relative weight to apply to the mask loss
        endpoint_loss_weight : float, default 1.
            Relative weight to apply to the endpoint classification
        return_mask_labels : bool, default False
            If `True`, returns the masks used to compute the mask loss
        restrict_to_clusters : bool, default False
            If `True`, when computing the positive labels for PPN, it will only
            look for points that are close to a given PPN label point with the
            same particle id.
        """
        # Store the loss parameters
        self.resolution = resolution
        self.balance_mask_loss = balance_mask_loss
        self.mask_weighting_mode = mask_weighting_mode
        self.balance_type_loss = balance_type_loss
        self.type_weighting_mode = type_weighting_mode
        self.reg_loss_weight = reg_loss_weight
        self.type_loss_weight = type_loss_weight
        self.mask_loss_weight = mask_loss_weight
        self.endpoint_loss_weight = endpoint_loss_weight
        self.return_mask_labels = return_mask_labels
        self.restrict_to_clusters = restrict_to_clusters
        self.point_classes = point_classes
        if point_classes is not None and isinstance(point_classes, int):
            self.point_classes = [point_classes]

        # Instantiate the regression loss function
        self.reg_loss_fn = torch.nn.MSELoss(reduction="mean")

        # Instantiate the point type loss function
        self.type_loss_fn = torch.nn.functional.cross_entropy

        # Instantiate the end point tyepe loss function
        self.end_loss_fn = torch.nn.functional.cross_entropy

        # Instantiate the mask loss function
        self.mask_loss = mask_loss
        if mask_loss == "CE":
            self.mask_loss_fn = torch.nn.functional.cross_entropy
        else:
            raise ValueError(f"Mask loss name not recognized: {mask_loss}")

    @staticmethod
    def get_ppn_positives(
        coords: torch.Tensor,
        ppn_labels: torch.Tensor,
        resolution: float,
        offset: int,
        labels: torch.Tensor = None,
    ):
        """Get ppn positive label mask.

        If the voxel `labels` are provided, they are used to restrict the mask
        applied to voxels within some distance of label points associated with
        the correct particle instance, not any particle instance.

        Parameters
        ----------
        coords : torch.Tensor
            (N, 3) 3D coordinates of the image voxels
        ppn_label : torch.Tensor
            (N, 1 + D + N_l) Tensor of PPN labels for the batch
        resolution : float
            Distance from a label point in pixels within which a voxel is
            considered positive (pixel of interest)
        offset : int
            The index offset needed to transform within-batch index to
            global (image) index.
        labels : torch.Tensor, optional
            (N) tensor of the particle id label for each voxel

        Returns
        -------
        positives : torch.Tensor
            (N) tensor of the positive label mask
        closests : torch.Tensor
            (N) tensor of the closest label point index
        """
        # Detach this process from the computation graph (mask not learnable)
        with torch.no_grad():
            # Compute the distance from the PPN labels to all the image points
            dist_mat = cdist_fast(ppn_labels[:, COORD_COLS], coords)

            # Mask out particle voxels for which the particle ID disagrees
            if labels is not None:
                bad_mask = ppn_labels[:, [PPN_LPART_COL]] != labels
                dist_mat[bad_mask] = torch.inf

            # Generate a positive mask for all particle voxels within some
            # distance of their label points
            positives = (dist_mat < resolution).any(dim=0)

            # Assign the closest label point to each postive particle voxel
            pos_index = torch.where(positives)[0]
            closests = torch.full(
                (len(coords),), -1, dtype=torch.long, device=coords.device
            )
            closests[pos_index] = offset + torch.argmin(dist_mat[:, pos_index], dim=0)

            return positives, closests

    def forward(
        self,
        ppn_label,
        ppn_points,
        ppn_masks,
        ppn_layers,
        ppn_coords,
        ppn_output_coords,
        ppn_classify_endpoints=None,
        clust_label=None,
        **kwargs,
    ):
        """Computes the three PPN losses.

        Parameters
        ----------
        ppn_label : TensorBatch
            (N, 1 + D + N_l) Tensor of PPN labels for the batch
        ppn_points : TensorBatch
            Complete PPN predictions at the last layer
        ppn_masks : List[TensorBatch]
            Binary mask at each layer of the PPN
        ppn_layers : List[TensorBatch]
            Output logits at each layer of the PPN
        ppn_coords : List[TensorBatch]
            Set of coordinates at each layer of the PPN
        ppn_output_coords : TensorBatch
            Set of coordinates at the very last layer of the PPN
        ppn_classify_endpoins : TensorBatch, optional
            Set of logits associated with end point classification
        clust_label : TensorBatch, optional
            (N, 1 + D + N_c) Tensor of cluster labels
            - N_c is is the number of cluster labels
        **kwargs : dict, optional
            Other outputs of the upstream model which are not relevant here

        Returns
        -------
        dict
            Dictionary of accuracies and losses
        """
        # Initialize the basics
        num_layers = len(ppn_layers)
        batch_size = ppn_label.batch_size

        # If requested, narrow down the list of label points
        if self.point_classes is not None:
            assert len(
                self.point_classes
            ), "Should provide at least one class to include in the loss"
            ppn_label_list = []
            for b, label_tensor in enumerate(ppn_label.split()):
                labels = label_tensor[:, PPN_LTYPE_COL]
                mask = torch.zeros(len(labels), dtype=torch.bool, device=labels.device)
                for c in self.point_classes:
                    mask |= labels == c
                valid_index = torch.where(mask)[0]
                ppn_label_list[b] = label_tensor[valid_index]

            ppn_label = TensorBatch.from_list(ppn_label_list)

        # Compute the label mask for the final PPN layer. Record which
        # label point is closest to each image voxel (defines label for it)
        coords_final = ppn_coords[-1]
        closest_list, positive_list = [], []
        part_labels = None
        offset = 0
        for b in range(batch_size):
            # If there are no label points, there are no positive points
            points_label = ppn_label[b]
            if not len(points_label):
                positive = torch.zeros(
                    coords_final.counts[b], dtype=torch.bool, device=coords_final.device
                )
                closest = torch.empty_like(positive, dtype=torch.long)
                positive_list.append(positive)
                closest_list.append(closest)
                continue

            # If needed, find which particle instance voxels belong to
            if self.restrict_to_clusters:
                assert clust_label is not None, (
                    "When using 'restrict_to_clusters', must provide "
                    "'clust_label' to the PPN loss."
                )
                part_labels = clust_label[b][:, PART_COL]

            # Assign positive/negative labels to each voxel in the image
            points_entry = coords_final[b][:, COORD_COLS] + 0.5
            positive, closest = self.get_ppn_positives(
                points_entry,
                points_label,
                resolution=self.resolution,
                offset=offset,
                labels=part_labels,
            )

            # Append
            positive_list.append(positive)
            closest_list.append(closest)
            offset += len(points_label)

        closests = torch.cat(closest_list, dim=0)
        positives = torch.cat(positive_list, dim=0).long()

        # Downsample the final mask to each PPN layer, apply mask loss
        downsample = ME.MinkowskiMaxPooling(2, 2, dimension=3)  # TODO
        mask_tensor = ME.SparseTensor(
            positives[:, None].float(), coordinates=coords_final.tensor[:, :VALUE_COL]
        )

        dtype, device = ppn_label.tensor.dtype, ppn_label.tensor.device
        mask_losses = torch.zeros(num_layers, dtype=dtype, device=device)
        mask_accs = torch.zeros(num_layers, dtype=dtype, device=device)
        mask_label_list = []
        for i in range(0, num_layers):
            # Narrow down outputs to this specific layer
            layer = num_layers - 1 - i
            coords_layer = ppn_coords[layer]
            scores_layer = ppn_layers[layer]
            mask_labels = mask_tensor.F.flatten().long()

            # If requested, store the label features in a list
            if self.return_mask_labels:
                mask_label_list.append(
                    TensorBatch(mask_labels[:, None], coords_layer.counts)
                )

            # Compute the mask weights over the whole batch, if requested
            mask_weight = None
            if self.balance_mask_loss:
                mask_weight = get_class_weights(
                    mask_labels, 2, self.mask_weighting_mode
                )

            # Compute the mask loss for this layer, increment
            mask_losses[layer] = self.mask_loss_fn(
                scores_layer.tensor, mask_labels, weight=mask_weight, reduction="mean"
            )

            # Compute the mask accuracy for this layer/batch, increment
            with torch.no_grad():
                num_points = len(scores_layer.tensor)
                mask_pred = torch.argmax(scores_layer.tensor, dim=1)
                mask_accs[layer] = (mask_pred == mask_labels).sum() / num_points

            # Update the mask label for the next iteration
            if layer != 0:
                mask_tensor = downsample(mask_tensor)

        # Apply the other losses to the last layer only
        zero = torch.tensor(0.0, dtype=ppn_label.dtype, device=ppn_label.device)
        one = torch.tensor(1.0, dtype=ppn_label.dtype, device=ppn_label.device)
        type_loss, reg_loss, end_loss = zero, zero, zero
        type_acc, end_acc = one, one
        pos_mask = torch.where(positives)[0]
        if len(pos_mask):
            # Narrow the loss down to the true positive pixels
            # TODO: should this be predicted positive pixels?

            # Closest ppn point label (index) to given positive point
            closests = closests[pos_mask]

            anchors = coords_final.tensor[:, COORD_COLS] + 0.5
            pixel_pos = ppn_points.tensor[:, PPN_ROFF_COLS] + anchors
            pixel_scores = ppn_points.tensor[:, PPN_RPOS_COLS]
            pixel_logits = ppn_points.tensor[:, PPN_RTYPE_COLS]

            pixel_pos = pixel_pos[pos_mask]
            pixel_scores = pixel_scores[pos_mask]
            pixel_logits = pixel_logits[pos_mask]

            ########## Type loss ##########
            # Compute type weights over the whole batch, if requested
            type_labels = ppn_label.tensor[:, PPN_LTYPE_COL].long()
            type_weight = None
            if self.balance_type_loss:
                num_classes = pixel_logits.shape[1]
                type_weight = get_class_weights(
                    type_labels, num_classes, self.type_weighting_mode
                )

            # Compute the type loss
            # TODO: deal with having multiple valid labels (track/shower/etc.)
            closest_type_labels = type_labels[closests]
            type_loss = self.type_loss_fn(
                pixel_logits, closest_type_labels, weight=type_weight
            )

            # Compute the type accuracy
            with torch.no_grad():
                num_points = len(closest_type_labels)
                type_pred = torch.argmax(pixel_logits, dim=1)
                type_acc = (type_pred == closest_type_labels).sum() / num_points

            ########## Regression loss ##########
            # Compute the regression loss. The offset should point to
            # the closest label point from that voxel
            point_labels = ppn_label.tensor[:, COORD_COLS]
            closest_point_labels = point_labels[closests]
            reg_loss = self.reg_loss_fn(pixel_pos, closest_point_labels)

            ########## End points loss ##########
            # If the upstream models produced endpoint predictions, apply loss.
            # Narrow the problem down to predictions closest to track points
            track_index = torch.where(closest_type_labels == TRACK_SHP)[0]
            if ppn_classify_endpoints and len(track_index):
                # Get the end point predictions
                end_logits = ppn_classify_endpoints.tensor[pos_mask]
                end_logits = end_logits[track_index]

                # Compute the end point classification loss
                # TODO: deal with having multiple valid labels (start/end)
                end_labels = ppn_label.tensor[:, PPN_LENDP_COL].long()
                closest_end_labels = end_labels[closests]
                closest_end_labels = closest_end_labels[track_index]
                end_loss = self.end_loss_fn(end_logits, closest_end_labels)

                # Compute the end point classification accuracy
                with torch.no_grad():
                    num_points = len(closest_end_labels)
                    end_pred = torch.argmax(end_logits, dim=1)
                    end_acc = (end_pred == closest_end_labels).sum() / num_points

        # Combine the losses and accuracies
        mask_loss = torch.mean(mask_losses)
        mask_acc = torch.mean(mask_accs)

        loss = (
            self.mask_loss_weight * mask_loss
            + self.type_loss_weight * type_loss
            + self.reg_loss_weight * reg_loss
        )
        accuracy = (mask_acc + type_acc) / 2

        if ppn_classify_endpoints is not None:
            loss += self.endpoint_loss_weight * end_loss
            accurary = accuracy * 2.0 + end_acc

        # Prepare the result dictionary
        result = {
            "loss": loss,
            "accuracy": accuracy.item(),
            "mask_loss": mask_loss.item(),
            "mask_accuracy": mask_acc.item(),
            "type_loss": type_loss.item(),
            "type_accuracy": type_acc.item(),
            "reg_loss": reg_loss.item(),
        }

        # Add the endpoint metrics if present
        if ppn_classify_endpoints is not None:
            result["classify_endpoints_loss"] = end_loss.item()
            result["classify_endpoints_accuracy"] = end_acc.item()

        # Add the mask loss at each layer
        for layer in range(num_layers):
            result[f"mask_loss_layer_{layer}"] = mask_losses[layer]
            result[f"mask_accuracy_layer_{layer}"] = mask_accs[layer]

        # If needed, return the mask labels
        if self.return_mask_labels:
            result["mask_labels"] = mask_label_list[::-1]

        return result


class ExpandAs(nn.Module):
    """Expands a one dimensional feature tensor to a higher dimension.

    Given a sparse tensor with one dimensional features, expand the
    feature map to a given shape and return a newly constructed
    ME.SparseTensor. This is used to expand a score array and apply
    it to the entire feature tensor of the the input.
    """

    def forward(
        self, x, shape, propagate_all=False, use_binary_mask=False, score_threshold=0.5
    ):
        """Expand a tensor to the appropriate shape

        Parameters
        ----------
        x : torch.Tensor
            (N, 2) Input tensor
        shape : ntuple
            (N, X) Shape to expand the mask to
        propagate_all : bool, default False
            If `True`, sets all features to 1.
        use_binary_mask : bool, default False
            If `True`, sets all features to either 0 or 1
        scrore_threshold : float, default 0.5
            If `use_binary_mask == True`, sets the threshold above which
            the feature is 1.0 and below which it is 0.0
        """
        # If all features must be propagated, set all scores to 1.0
        assert x.F.shape[1] == 2, "Expects a two-score tensor"
        features = x.F[:, 1:]
        if propagate_all:
            features[...] = 1.0

        # Expand the features to the right dimension
        if use_binary_mask:
            features = (features > score_threshold).expand(*shape)
        else:
            features = features.expand(*shape)

        return ME.SparseTensor(
            features=features,
            coordinate_map_key=x.coordinate_map_key,
            coordinate_manager=x.coordinate_manager,
        )


class AttentionMask(torch.nn.Module):
    """Returns a masked tensor of x according to mask, where the number of
    coordinates between x and mask differ.
    """

    def __init__(self, score_threshold=0.5):
        """Initialize attention mask.

        Parameters
        ----------
        score_threshold : float, default 0.5
            Score above which a voxel is considered positive
        """
        # Initialize parent class
        super().__init__()

        # Pruning layer
        self.prune = ME.MinkowskiPruning()

        # Store parameters
        self.score_threshold = score_threshold

    def forward(self, x, mask):
        """Prune the input data.

        Parameters
        ----------
        x : ME.SparseTensor
            Input sparse tensor
        mask : ME.SparseTensor
            Mask to apply
        """
        assert x.tensor_stride == mask.tensor_stride

        device = x.F.device
        # Create a mask sparse tensor in x-coordinates
        x0 = ME.SparseTensor(
            coordinates=x.C,
            features=torch.zeros(x.F.shape[0], mask.F.shape[1]).to(device),
            coordinate_manager=x.coordinate_manager,
            tensor_stride=x.tensor_stride,
        )

        mask_in_xcoords = x0 + mask

        x_expanded = ME.SparseTensor(
            coordinates=mask_in_xcoords.C,
            features=torch.zeros(mask_in_xcoords.F.shape[0], x.F.shape[1]).to(device),
            coordinate_manager=x.coordinate_manager,
            tensor_stride=x.tensor_stride,
        )

        x_expanded = x_expanded + x

        target = mask_in_xcoords.F.int().bool().squeeze()
        x_pruned = self.prune(x_expanded, target)
        return x_pruned


class MergeConcat(torch.nn.Module):
    """Merge one sparse tensor with another."""

    def forward(self, x, other):
        """Merge two sparse tensors

        Parameters
        ----------
        x : ME.SparseTensor
            Input sparse tensor
        other : ME.SparseTensor
            Other sparse tensor to merge

        Returns
        -------
        ME.SparseTensor
            Concatenated sparse tensor
        """
        assert x.tensor_stride == other.tensor_stride
        device = x.F.device

        # Create a placeholder tensor with x.C coordinates
        x0 = ME.SparseTensor(
            coordinates=x.C,
            features=torch.zeros(x.F.shape[0], other.F.shape[1]).to(device),
            coordinate_manager=x.coordinate_manager,
            tensor_stride=x.tensor_stride,
        )

        # Set placeholder values with other.F features by performing
        # sparse tensor addition.
        x1 = x0 + other

        # Same procedure, but with other
        x_expanded = ME.SparseTensor(
            coordinates=x1.C,
            features=torch.zeros(x1.F.shape[0], x.F.shape[1]).to(device),
            coordinate_manager=x.coordinate_manager,
            tensor_stride=x.tensor_stride,
        )

        x2 = x_expanded + x

        # Now x and other share the same coordinates and shape
        concated = ME.cat(x1, x2)
        return concated


class MinkGhostMask(torch.nn.Module):
    """Ghost mask downsampler.

    Downsamples the ghost mask and prunes a tensor with current
    ghost mask to obtain nonghost tensor and the new ghost mask.
    """

    def __init__(self):
        """Initialize the mask downsampler."""
        # Initialize parent class
        super().__init__()

        # Initialize the downsampler
        self.downsample = ME.MinkowskiMaxPooling(2, 2, dimension=3)

        # Set the layer in evaluation mode (no gradients)
        self.eval()

    def forward(self, ghost_mask, premask_tensor):
        """Applies mask and downsamples it for the next layer.

        Parameters
        ----------
        ghost_mask : ME.SparseTensor
            Current resolution ghost mask
        premask_tensor : ME.SparseTensor
            Current resolution feature map to be pruned

        Returns
        -------
        downsampled_mask : ME.SparseTensor)
            2x2 downsampled ghost mask
        downsampled_tensor : ME.SparseTensor
            2x2 downsampled feature map
        """
        # assert ghost_mask.shape[0] == premask_tensor.shape[0]
        with torch.no_grad():
            factor = premask_tensor.tensor_stride[0]
            for i in range(np.log2(factor).astype(int)):
                ghost_mask = self.downsample(ghost_mask)

            return ghost_mask



import time
from dataclasses import dataclass


@dataclass
class Time:
    """Simple dataclass to hold time information.

    Attributes
    ----------
    wall : float, optional
         Wall time
    cpu : float, optional
         CPU time
    """

    wall: float = None
    cpu: float = None

    def __add__(self, time):
        """Overload the addition operator.

        Parameters
        ----------
        time : Time
            Other Time object

        Returns
        -------
        Time
           Summed times
        """
        return Time(wall=self.wall + time.wall, cpu=self.cpu + time.cpu)

    def __sub__(self, time):
        """Overload the substraction operator.

        Parameters
        ----------
        time : Time
            Other Time object

        Returns
        -------
        Time
           Substracted times
        """
        return Time(wall=self.wall - time.wall, cpu=self.cpu - time.cpu)

    def __eq__(self, time):
        """Overload the equality operator.

        Parameters
        ----------
        time : Time
            Other Time object

        Returns
        -------
        bool
            True if both times are identical
        """
        if isinstance(time, type(self)):
            return self.wall == time.wall and self.cpu == time.cpu
        else:
            return self.wall == time and self.cpu == time

    def copy(self):
        """Returns an independant copy of the object.

        Returns
        -------
        Time
            Copy of the object
        """
        return Time(wall=self.wall, cpu=self.cpu)

    @classmethod
    def current(cls):
        """Simple function which returns the current time (wall and cpu).

        Returns
        -------
        Time
           Current time
        """
        return cls(time.time(), time.process_time())


class Stopwatch:
    """Simple class to hold timing information for a specific process."""

    def __init__(self):
        """Give default values to the underlying class attributes."""
        self._start = Time()
        self._stop = Time()
        self._pause = Time()
        self._time = Time(0.0, 0.0)
        self._total = Time(0.0, 0.0)

    @property
    def running(self):
        """Whether the stopwatch is currently running."""
        return self._start != Time() and self._stop == Time()

    @property
    def paused(self):
        """Whether the stopwatch is currently paused."""
        return self._pause != Time() and self._stop == Time()

    @property
    def start(self):
        """Time when the stopwatch was last started."""
        return self._start

    @start.setter
    def start(self, start):
        # Check that the watch was not already started
        if self.running:
            raise ValueError("Cannot restart a watch that has not been stopped.")

        # Start watch, reinitialize stop
        self._start = start
        self._stop = Time()
        if self._pause == Time():
            self._time = Time(0.0, 0.0)

    @property
    def stop(self):
        """Time when the stopwatch was last stopped."""
        return self._stop

    @stop.setter
    def stop(self, stop):
        # Check that the watch was started
        if self._start == Time():
            raise ValueError("Cannot stop a watch that has not been started.")

        # Check that the watch was not already stopped
        if self._stop != Time():
            raise ValueError("Cannot stop a watch more than once.")

        # Stop the watch, record the relevant quantities
        self._stop = stop
        self._pause = Time()
        self._time += self.stop - self.start
        self._total += self.time

    @property
    def pause(self):
        """Time when the stopwatch was last paused."""
        return self._pause

    @pause.setter
    def pause(self, pause):
        # Check that the watch was started
        if self._start == Time():
            raise ValueError("Cannot pause a watch that has not been started.")

        # Check that the watch was not already stopped
        if self._stop != Time():
            raise ValueError("Cannot pause a watch that has been stopped.")

        # Increment the time, reset the start
        self._pause = pause
        self._time += self.pause - self.start
        self._start = Time()

    @property
    def time(self):
        """Time between the last start and the last stop."""
        # Check that the watch was stopped
        if self._stop == Time():
            raise ValueError("Cannot get time of watch that has not been stopped.")

        return self._time

    @property
    def time_sum(self):
        """Sum of times between all watch starts en stops."""
        # Check that the watch was stopped
        if self._stop == Time():
            raise ValueError("Cannot get time of watch that has not been stopped.")

        return self._total


class StopwatchManager:
    """Simple class to organize various time measurements."""

    def __init__(self):
        """Initalize the basic private stopwatch attributes."""
        self._watch = {}

    def keys(self):
        """Get the list of all initialized stopwatch tags.

        Returns
        -------
        List[str]
            List of stopwatch names
        """
        return self._watch.keys()

    def values(self):
        """Get the list of all initialized stopwatches.

        Returns
        -------
        List[Stopwatch]
            List of stopwatch objects
        """
        return self._watch.values()

    def items(self):
        """Get the list of all initialized stopwatch tags and the
        corresponding Stopwatch object for each of them.

        Returns
        -------
        List[Tuple[str, Stopwatch]]
            List of (key, stopwatch) pairs
        """
        return self._watch.items()

    def initialize(self, key):
        """Initialize one stopwatch. If it's already been initialized,
        reset the global counters to 0.

        Parameters
        ----------
        key : Union[str, List[str]]
            Key or list of keys to initialize a `Stopwatch` for
        """
        # Loop over keys
        keys = [key] if isinstance(key, str) else key
        for k in keys:
            # Initialize stopwatch
            self._watch[k] = Stopwatch()

    def reset(self, key=None):
        """Reset a stopwatch to its initial state.

        Parameters
        ----------
        key : Union[str, List[str]], optional
            Key or list of keys to reset a `Stopwatch` for. If None, reset all stopwatches.
        """
        # Get the list of keys to reset. If not specified, reset all
        key = self.keys() if key is None else key

        # Loop over keys
        keys = [key] if isinstance(key, str) else key
        for k in keys:
            # Check that a stopwatch exists
            if not k in self._watch:
                raise KeyError(f"No stopwatch initialized under the name: {k}")

            # Reset stopwatch
            self._watch[k] = Stopwatch()

    def start(self, key):
        """Starts a stopwatch for a unique key.

        Parameters
        ----------
        key : Union[str, List[str]]
            Key or list of keys for which to start the clock
        """
        # Loop over keys
        start_time = Time.current()
        keys = [key] if isinstance(key, str) else key
        for k in keys:
            # If this is the first time, initialize a new Stopwatch
            if not k in self._watch:
                raise KeyError(f"No stopwatch initialized under the name: {k}")

            # Reinitialize the watch
            self._watch[k].start = start_time.copy()

    def stop(self, key):
        """Stops a stopwatch for a unique key.

        Parameters
        ----------
        key : str
            Key for which to stop the clock
        """
        # Loop over keys
        stop_time = Time.current()
        keys = [key] if isinstance(key, str) else key
        for k in keys:
            # Check that a stopwatch exists
            if not k in self._watch:
                raise KeyError(f"No stopwatch started under the name: {k}")

            # Stop
            self._watch[k].stop = stop_time.copy()

    def pause(self, key):
        """Temporarily pause a watch for a unique key.

        Parameters
        ----------
        key : str
            Key for which to pause the clock
        """
        # Loop over keys
        pause_time = Time.current()
        keys = [key] if isinstance(key, str) else key
        for k in keys:
            # Check that a stopwatch exists
            if not k in self._watch:
                raise KeyError(f"No stopwatch started under the name: {k}")

            # Stop
            self._watch[k].pause = pause_time.copy()

    def time(self, key):
        """Returns the time recorded since the last start.

        Parameters
        ----------
        key : str
            Key for which to return the time

        Returns
        -------
        Time
            Execution time of one iteration of a process
        """
        # Check that a stopwatch exists
        if not key in self._watch:
            raise KeyError(f"No stopwatch started under the name: {key}")

        # Return the time since the start
        return self._watch[key].time

    def time_sum(self, key):
        """Returns the sum of times recorded between each start/stop pairs.

        Parameters
        ----------
        key : str
            Key for which to return the time

        Returns
        -------
        Time
            Execution time of all iterations of a process so far
        """
        # Check that a stopwatch exists
        if not key in self._watch:
            raise KeyError(f"No stopwatch started under the name: {key}")

        # Return the time since the start
        return self._watch[key].time_sum

    def times(self):
        """Returns the times for each of the stopwatches as a dictionary.

        Returns
        -------
        Dict[str, Time]
            Execution time of one iteration of each process
        """
        return {key: value.time for key, value in self.items()}

    def times_sum(self):
        """Returns the times for each of the stopwatches as a dictionary.

        Returns
        -------
        Dict[str, Time]
            Execution time of all iterations of each process so far
        """
        return {key: value.time_sum for key, value in self.items()}

    def update(self, other, prefix=None):
        """Updates this manager with values from another stopwatch manager.

        Parameters
        ----------
        other : StopwatchManager
             Dictionary of execution times from another process
        prefix : str, optional
             String to prefix the timer key with

        Returns
        -------
        Dict[str, Time]
            Combined execution time of all iterations of each process so far
        """
        for key, value in other.items():
            if prefix is None:
                self._watch[key] = value
            else:
                self._watch[f"{prefix}_{key}"] = value



"""Particle identification modules."""

import numpy as np

from spine.post.base import PostBase
from spine.utils.globals import TRACK_SHP
from spine.utils.pid import TemplateParticleIdentifier

__all__ = ["PIDTemplateProcessor"]


class PIDTemplateProcessor(PostBase):
    """Produces particle species classification estimates based on dE/dx vs
    residual range templates of tracks.
    """

    # Name of the post-processor (as specified in the configuration)
    name = "pid_template"

    def __init__(
        self,
        fill_per_pid=False,
        obj_type="particle",
        run_mode="reco",
        truth_point_mode="points",
        truth_dep_mode="depositions",
        **identifier,
    ):
        """Store the necessary attributes to do template-based PID prediction.

        Parameters
        ----------
        fill_per_pid : bool, default False
            If `True`, stores the scores associated with each PID candidate
        **identifier : dict, optional
            Particle template identifier configuration parameters
        """
        # Initialize the parent class
        super().__init__(obj_type, run_mode, truth_point_mode, truth_dep_mode)

        # Store additional parameter
        self.fill_per_pid = fill_per_pid

        # Initialize the underlying template-fitter class
        self.identifier = TemplateParticleIdentifier(**identifier)

    def process(self, data):
        """Reconstruct the CSDA KE estimates for each particle in one entry.

        Parameters
        ----------
        data : dict
            Dictionary of data products
        """
        # Loop over particle objects
        for k in self.fragment_keys + self.particle_keys:
            for obj in data[k]:
                # Only run this algorithm on tracks that have a CSDA table
                if not obj.shape == TRACK_SHP:
                    continue

                # Make sure the object coordinates are expressed in cm
                self.check_units(obj)

                # Get point coordinates and depositions
                points = self.get_points(obj)
                values = self.get_depositions(obj)
                if len(points) == 0:
                    continue

                # Run the particle identifier
                pid, chi2_values = self.identifier(
                    points, values, obj.end_point, obj.start_point
                )

                # Store for this PID
                obj.chi2_pid = pid
                if self.fill_per_pid:
                    chi2_per_pid = np.full(
                        len(obj.chi2_per_pid), -1.0, dtype=obj.chi2_per_pid.dtype
                    )
                    for i, pid in enumerate(self.identifier.include_pids):
                        chi2_per_pid[pid] = chi2_values[i]

                    obj.chi2_per_pid = chi2_per_pid



"""I/O parsers are used to read data products from a LArCV ROOT file.

Parsers are listed under :mod:`io.dataset.schema` in the configuration.
`schema` is a list of named values. Each name is arbitrary and will be
used as a key to access the output of the parser in a dictionary.

List of existing parsers
------------------------

.. csv-table:: Sparse parsers
    :header: Parser name, Description

    ``sparse2d``, Retrieve sparse tensor input from
    larcv::EventSparseTensor2D object
    ``sparse3d``, Retrieve sparse tensor input from
    larcv::EventSparseTensor3D object
    ``sparse3d_ghost``, Takes semantics tensor and turns its labels into
    ghost labels

.. csv-table:: Cluster parsers
    :header: Parser name, Description

    ``cluster2d``, Retrieve list of sparse tensor input from
    larcv::EventClusterPixel2D
    ``cluster3d``, Retrieve list of sparse tensor input from
    larcv::EventClusterVoxel3D

.. csv-table:: Particle parsers
    :header: Parser name, Description

    ``particle``, Retrieve array of larcv::Particle
    ``neutrino``, Retrieve array of larcv::Neutrino
    ``particle_points``, Retrieve array of larcv::Particle ground truth
    points tensor
    ``particle_coords``, Retrieve array of larcv::Particle coordinates
    (start and end) and start time
    ``particle_graph``, Construct edges between particles (i.e. clusters)
    from larcv::EventParticle
    ``single_particle_pid``, Get a single larcv::Particle PDG code
    ``single_particle_energy``, Get a single larcv::Particle initial
    energy

.. csv-table:: Miscellaneous parsers
    :header: Parser name, Description

    ``meta``, Get the meta information to translate into real world coordinates
    ``run_info``, Parse run info (run, subrun, event number)
    ``flash``, Parse optical flashes
    ``crthit``, Parse cosmic ray tagger hits
    ``trigger``, Parse trigger information


What does a typical parser configuration look like?
---------------------------------------------------
If the configuration looks like this, for example:

..  code-block:: yaml

    schema:
      input_data:
        parser: sparse3d
        sevent_list:
          - sparse3d_reco
          - sparse3d_reco_chi2

Where `input_data` is an arbitrary name chosen by the user, which will be the
key to access the output of the parser `sparse3d`. The parser arguments
can be ROOT TTree names that will be fed to the parser or parser arguments. The
arguments must be passed as a dictionary of (argument name, value) pairs. In
this example, the parser will be called with a list of 2 objects: A
:class:`larcv::EventSparseTensor3D` coming from the ROOT TTree `sparse3d_reco`,
and another one coming from the TTree `sparse3d_reco_chi2`.

How do I know what a parser requires?
-------------------------------------
To be completed.

How do I know what my ROOT file contains?
-----------------------------------------
To be completed.
"""

from .cluster import *
from .misc import *
from .particle import *
from .sparse import *



"""Module with a data class object which represents true particle information.

This copies the internal structure of :class:`larcv.Particle`.
"""

from dataclasses import dataclass, field
from warnings import warn

import numpy as np

from spine.utils.globals import PID_LABELS, SHAPE_LABELS, UNKWN_SHP

from .base import PosDataBase

__all__ = ["Particle"]


@dataclass(eq=False)
class Particle(PosDataBase):
    """Particle truth information.

    Attributes
    ----------
    id : int
        Index of the particle in the list
    mct_index : int
        Index in the original MCTruth array from whence it came
    mcst_index : int
        Index in the original MCTrack/MCShower array from whence it came
    group_id : int
        Index of the group the particle belongs to
    interaction_id : int
        Index of the interaction the partile belongs to
    nu_id : int
        Index of the neutrino this particle belongs to
    interaction_primary : int
        Whether the particle is primary in its interaction or not
    group_primary : int
        Whether this particle is primary in its group or not
    parent_id : int
        Index of the parent particle
    children_id : np.ndarray
        List of indexes of the children particles
    track_id : int
        Geant4 track ID
    parent_track_id : int
        Geant4 track ID of the parent particle
    ancestor_track_id : int
        Geant4 track ID of the ancestor particle
    shape : int
        Enumerated semantic type of the particle
    num_voxels : int
        Number of voxels matched to this particle instance
    energy_init : float
        True initial energy in MeV
    energy_deposit : float
        Amount of energy matched to this particle instance in MeV
    distance_travel : float
        True amount of distance traveled by the particle in the active volume
    creation_process : str
        Creation process
    parent_creation_process : str
        Creation process of the parent particle
    ancestor_creation_process : str
        Creation process of the ancestor particle
    pid : int
        Enumerated particle species type of the particle
    pdg_code : int
        Particle PDG code
    parent_pdg_code : int
        Particle PDG code of the parent particle
    ancestor_pdg_code : int
        Particle PDG code of the ancestor particle
    t : float
        Particle creation time (ns)
    end_t : float
        Particle death time (ns)
    parent_t : float
        Particle creation time of the parent particle (ns)
    ancestor_t : float
        Particle creation time of the ancestor particle (ns)
    position : np.ndarray
        Location of the creation point of the particle
    end_position : np.ndarray
        Location where the particle stopped
    parent_position : np.ndarry
        Location of the creation point of the parent particle
    ancestor_position : np.ndarray
        Location of the creation point of the ancestor particle
    first_step : np.ndarray
        Location of the first energy deposition of the particle
    last_step : np.ndarray
        Location of the last energy deposition of the particle
    momentum : np.ndarray
        3-momentum of the particle at the production point
    end_momentum : np.ndarray
        3-momentum of the particle at where it stops or exits the detector
    p : float
        Momentum magnitude of the particle at the production point
    end_p : float
        Momentum magnitude of the particle where it stops or exits the detector
    mass : float
        Rest mass of the particle in MeV/c^2
    units : str
        Units in which the position attributes are expressed
    """

    # Attributes
    id: int = -1
    mct_index: int = -1
    mcst_index: int = -1
    group_id: int = -1
    interaction_id: int = -1
    nu_id: int = -1
    interaction_primary: int = -1
    group_primary: int = -1
    parent_id: int = -1
    children_id: np.ndarray = None
    track_id: int = -1
    parent_track_id: int = -1
    ancestor_track_id: int = -1
    pid: int = -1
    pdg_code: int = -1
    parent_pdg_code: int = -1
    ancestor_pdg_code: int = -1
    num_voxels: int = -1
    shape: int = UNKWN_SHP
    energy_init: float = -1.0
    energy_deposit: float = -1.0
    distance_travel: float = -1.0
    creation_process: str = ""
    parent_creation_process: str = ""
    ancestor_creation_process: str = ""
    t: float = -np.inf
    end_t: float = -np.inf
    parent_t: float = -np.inf
    ancestor_t: float = -np.inf
    position: np.ndarray = None
    end_position: np.ndarray = None
    parent_position: np.ndarray = None
    ancestor_position: np.ndarray = None
    first_step: np.ndarray = None
    last_step: np.ndarray = None
    momentum: np.ndarray = None
    end_momentum: np.ndarray = None
    mass: float = None
    p: float = None
    end_p: float = None
    units: str = "cm"

    # Fixed-length attributes
    _fixed_length_attrs = (
        ("position", 3),
        ("end_position", 3),
        ("parent_position", 3),
        ("ancestor_position", 3),
        ("first_step", 3),
        ("last_step", 3),
        ("momentum", 3),
        ("end_momentum", 3),
    )

    # Variable-length attributes
    _var_length_attrs = (("children_id", np.int64),)

    # Attributes specifying coordinates
    _pos_attrs = (
        "position",
        "end_position",
        "parent_position",
        "ancestor_position",
        "first_step",
        "last_step",
    )

    # Attributes specifying vector components
    _vec_attrs = ("momentum", "end_momentum")

    # Enumerated attributes
    _enum_attrs = (
        ("shape", tuple((v, k) for k, v in SHAPE_LABELS.items())),
        ("pid", tuple((v, k) for k, v in PID_LABELS.items())),
    )

    # String attributes
    _str_attrs = (
        "creation_process",
        "parent_creation_process",
        "ancestor_creation_process",
    )

    # Index attributes
    _index_attrs = (
        "id",
        "parent_id",
        "group_id",
        "parent_id",
        "interaction_id",
        "nu_id",
        "children_id",
    )

    @property
    def p(self):
        """Computes the magnitude of the initial momentum.

        Returns
        -------
        float
            Norm of the initial momentum vector
        """
        return np.linalg.norm(self.momentum)

    @p.setter
    def p(self, p):
        pass

    @property
    def end_p(self):
        """Computes the magnitude of the final momentum.

        Returns
        -------
        float
            Norm of the final momentum vector
        """
        return np.linalg.norm(self.end_momentum)

    @end_p.setter
    def end_p(self, end_p):
        pass

    @property
    def mass(self):
        """Computes the rest mass of the particle from its energy/momentum.

        Returns
        -------
        float
            Rest mass of the particle in MeV/c^2
        """
        if self.energy_init < 0.0:
            return -1.0

        return np.sqrt(max(0.0, self.energy_init**2 - np.sum(self.momentum**2)))

    @mass.setter
    def mass(self, mass):
        pass

    @classmethod
    def from_larcv(cls, particle):
        """Builds and returns a Particle object from a LArCV Particle object.

        Parameters
        ----------
        particle : larcv.Particle
            LArCV-format particle object

        Returns
        -------
        Particle
            Particle object
        """
        # Initialize the dictionary to initialize the object with
        obj_dict = {}

        # Load the scalar attributes
        for prefix in ("", "parent_", "ancestor_"):
            for key in ("track_id", "pdg_code", "creation_process", "t"):
                obj_dict[prefix + key] = getattr(particle, prefix + key)()
        for key in (
            "id",
            "group_id",
            "interaction_id",
            "parent_id",
            "mct_index",
            "mcst_index",
            "num_voxels",
            "shape",
            "energy_init",
            "energy_deposit",
            "distance_travel",
        ):
            if not hasattr(particle, key):
                warn(
                    f"The LArCV Particle object is missing the {key} "
                    "attribute. It will miss from the Particle object."
                )
                continue
            obj_dict[key] = getattr(particle, key)()

        obj_dict["end_t"] = particle.end_position().t()

        # Load the positional attribute
        pos_attrs = ["x", "y", "z"]
        for key in cls._pos_attrs:
            vector = getattr(particle, key)()
            obj_dict[key] = np.asarray(
                [getattr(vector, a)() for a in pos_attrs], dtype=np.float32
            )

        # Load the other array attributes (special care needed). Note for future
        # self: need the list comprehension. Direct casting is INSANELY slow...
        obj_dict["children_id"] = np.asarray(
            [i for i in particle.children_id()], dtype=int
        )

        mom_attrs = ("px", "py", "pz")
        for prefix in ("", "end_"):
            key = prefix + "momentum"
            if not hasattr(particle, key):
                warn(
                    f"The LArCV Particle object is missing the {key} "
                    "attribute. It will miss from the Particle object."
                )
                continue
            obj_dict[key] = np.asarray(
                [getattr(particle, prefix + a)() for a in mom_attrs], dtype=np.float32
            )

        return cls(**obj_dict)



"""Image classification module.

This module includes:
    - Single full image classification
    - Individual cluster classification
    - UQ implementations of the full image classification
"""

from collections import Counter, OrderedDict, defaultdict

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch_geometric.data import Batch, Data

from spine.data import TensorBatch
from spine.utils.globals import PID_COL
from spine.utils.gnn.cluster import form_clusters, get_cluster_label

# from .experimental.layer.pointmlp import PointMLPEncoder
from .experimental.bayes.encoder import MCDropoutEncoder
from .experimental.bayes.evidential import EVDLoss
from .experimental.layer.pointnet import PointNetEncoder
from .layer.cnn.act_norm import act_factory
from .layer.cnn.configuration import setup_cnn_configuration
from .layer.cnn.encoder import SparseResidualEncoder
from .layer.factories import loss_fn_factory


class MultiParticleImageClassifier(ImageClassifier):

    MODULES = ["particle_image_classifier", "network_base", "mink_encoder"]

    def __init__(self, cfg, name="particle_image_classifier"):
        super(MultiParticleImageClassifier, self).__init__(cfg, name)

        model_cfg = cfg.get(name, {})
        self.batch_col = model_cfg.get("batch_col", 0)
        self.split_col = model_cfg.get("split_col", 6)
        self.num_classes = model_cfg.get("num_classes", 5)

        self.skip_invalid = model_cfg.get("skip_invalid", True)
        self.target_col = model_cfg.get("target_col", 9)
        self.invalid_id = model_cfg.get("invalid_id", -1)

        self.split_input_mode = model_cfg.get("split_input_as_tg_batch", False)

    def split_input_as_tg_batch(self, point_cloud, clusts=None):
        point_cloud_cpu = point_cloud.detach().cpu().numpy()
        batches, bcounts = np.unique(
            point_cloud_cpu[:, self.batch_col], return_counts=True
        )
        if clusts is None:
            clusts = form_clusters(point_cloud_cpu, column=self.split_col)
        if not len(clusts):
            return Batch()

        if self.skip_invalid:
            target_ids = get_cluster_label(
                point_cloud_cpu, clusts, column=self.target_col
            )
            clusts = [
                c for i, c in enumerate(clusts) if target_ids[i] != self.invalid_id
            ]
            if not len(clusts):
                return Batch()

        data_list = []
        for i, c in enumerate(clusts):
            x = point_cloud[c, 4].view(-1, 1)
            pos = point_cloud[c, 1:4]
            data = Data(x=x, pos=pos)
            data_list.append(data)

        split_data = Batch.from_data_list(data_list)
        return split_data, clusts

    def split_input(self, point_cloud, clusts=None):
        point_cloud_cpu = point_cloud.detach().cpu().numpy()
        batches, bcounts = np.unique(
            point_cloud_cpu[:, self.batch_col], return_counts=True
        )
        if clusts is None:
            clusts = form_clusters(point_cloud_cpu, column=self.split_col)
        if not len(clusts):
            return point_cloud, [np.array([]) for _ in batches], []

        if self.skip_invalid:
            target_ids = get_cluster_label(
                point_cloud_cpu, clusts, column=self.target_col
            )
            clusts = [
                c for i, c in enumerate(clusts) if target_ids[i] != self.invalid_id
            ]
            if not len(clusts):
                return point_cloud, [np.array([]) for _ in batches], []

        split_point_cloud = point_cloud.clone()
        split_point_cloud[:, self.batch_col] = -1
        for i, c in enumerate(clusts):
            split_point_cloud[c, self.batch_col] = i

        batch_ids = get_cluster_label(point_cloud_cpu, clusts, column=self.batch_col)
        clusts_split, cbids = split_clusts(clusts, batch_ids, batches, bcounts)

        return (
            split_point_cloud[split_point_cloud[:, self.batch_col] > -1],
            clusts_split,
            cbids,
        )

    def forward(self, input, clusts=None):
        res = {}
        (point_cloud,) = input

        # It is possible that pid = 5 appears in the 9th column.
        # In that case, it is observed that the training crashses with a
        # integer overflow numel error.
        mask = point_cloud[:, PID_COL] < self.num_classes
        valid_points = point_cloud[mask]

        if self.split_input_mode:
            batch, clusts = self.split_input_as_tg_batch(valid_points, clusts)
            out = self.encoder(batch)
            out = self.final_layer(out)
            res["clusts"] = [clusts]
            res["logits"] = [out]
        else:
            out, clusts_split, cbids = self.split_input(valid_points, clusts)
            res["clusts"] = [clusts_split]

            out = self.encoder(out)
            out = self.final_layer(out)
            res["logits"] = [[out[b] for b in cbids]]

        return res


class DUQParticleClassifier(ImageClassifier):
    """
    Uncertainty Estimation Using a Single Deep Deterministic Neural Network
    https://arxiv.org/pdf/2003.02037.pdf
    Joost van Amersfoort, Lewis Smith, Yee Whye Teh, Yarin Gal.

    Pytorch Implementation for SparseConvNets with MinkowskiEngine backend.
    """

    MODULES = ["network_base", "particle_image_classifier", "mink_encoder"]

    def __init__(self, cfg, name="duq_particle_classifier"):
        super(DUQParticleClassifier, self).__init__(cfg, name=name)
        self.model_config = cfg.get(name, {})
        self.final_layer = None
        self.gamma = self.model_config.get("gamma", 0.999)
        self.sigma = self.model_config.get("sigma", 0.3)

        self.embedding_dim = self.model_config.get("embedding_dim", 64)
        self.latent_size = self.model_config.get("latent_size", 256)

        self.w = nn.Parameter(
            torch.zeros(self.embedding_dim, self.num_classes, self.latent_size)
        )

        nn.init.kaiming_normal_(self.w, nonlinearity="relu")

        self.register_buffer("N", torch.ones(self.num_classes) * 20)
        self.register_buffer(
            "m", torch.normal(torch.zeros(self.embedding_dim, self.num_classes), 0.05)
        )

        self.m = self.m * self.N.unsqueeze(0)

    def embed(self, x):

        feats = self.encoder(x)
        out = torch.einsum("ij,mnj->imn", feats, self.w)
        return out

    def bilinear(self, z):
        embeddings = self.m / self.N.unsqueeze(0)

        diff = z - embeddings.unsqueeze(0)
        y_pred = (-(diff**2)).mean(1).div(2 * self.sigma**2).exp()

        return y_pred

    def forward(self, input):

        (point_cloud,) = input
        if self.training:
            point_cloud.requires_grad_(True)

        z = self.embed(point_cloud)
        y_pred = self.bilinear(z)

        res = {
            "score": [y_pred],
            "embedding": [z],
            "input": [point_cloud],
            "centroids": [
                self.m.detach().cpu().numpy() / self.N.detach().cpu().numpy()
            ],
        }

        self.z = z
        self.y_pred = y_pred

        return res

    def update_buffers(self):
        with torch.no_grad():
            # normalizing value per class, assumes y is one_hot encoded
            self.N = torch.max(
                self.gamma * self.N + (1 - self.gamma) * self.y_pred.sum(0),
                torch.ones_like(self.N),
            )
            # compute sum of embeddings on class by class basis
            features_sum = torch.einsum("ijk,ik->jk", self.z, self.y_pred)
            self.m = self.gamma * self.m + (1 - self.gamma) * features_sum


class EvidentialParticleClassifier(ImageClassifier):

    MODULES = ["network_base", "particle_image_classifier", "mink_encoder"]

    def __init__(self, cfg, name="evidential_image_classifier"):
        super(EvidentialParticleClassifier, self).__init__(cfg, name=name)
        self.final_layer_name = cfg.get(name, {}).get("final_layer_name", "relu")
        if self.final_layer_name == "relu":
            self.final_layer = nn.Sequential(
                nn.Linear(self.encoder.latent_size, self.num_classes), nn.ReLU()
            )
        elif self.final_layer_name == "softplus":
            self.final_layer = nn.Sequential(
                nn.Linear(self.encoder.latent_size, self.num_classes), nn.Softplus()
            )
        else:
            raise Exception(
                "Unknown output activation name %s provided" % self.final_layer_name
            )
        self.eps = cfg.get(name, {}).get("eps", 0.0)

    def forward(self, input):
        (point_cloud,) = input
        out = self.encoder(point_cloud)
        evidence = self.final_layer(out)
        # print("Evidence = ", evidence)
        concentration = evidence + 1.0
        S = torch.sum(concentration, dim=1, keepdim=True)
        uncertainty = self.num_classes / (S + self.eps)
        res = {
            "evidence": [evidence],
            "uncertainty": [uncertainty],
            "concentration": [concentration],
            "expected_probability": [concentration / S],
        }
        return res


class BayesianParticleClassifier(torch.nn.Module):

    MODULES = ["network_base", "mcdropout_encoder"]

    def __init__(self, cfg, name="bayesian_particle_classifier"):
        super(BayesianParticleClassifier, self).__init__()
        setup_cnn_configuration(self, cfg, "network_base")

        self.model_config = cfg.get(name, {})
        self.num_classes = self.model_config.get("num_classes", 5)
        self.encoder = MCDropoutEncoder(cfg)

        self.mode = self.model_config.get("mode", "mc_dropout")

        if self.mode == "evidential":
            self.logit_layer = nn.Sequential(
                nn.Linear(self.encoder.latent_size, self.num_classes), nn.Softplus()
            )
        else:
            self.logit_layer = nn.Sequential(
                nn.ReLU(), nn.Linear(self.encoder.latent_size, self.num_classes)
            )

        self.num_samples = self.model_config.get("num_samples", 20)
        self.eps = self.model_config.get("eps", 0.0)
        print("Dropout network will run inference on {} mode".format(self.mode))

    def evidential_forward(self, input):
        (point_cloud,) = input
        out = self.encoder(point_cloud)
        out = self.logit_layer(out) + self.eps

        concentration = out + 1.0
        S = torch.sum(concentration, dim=1, keepdim=True)
        uncertainty = self.num_classes / (S + 0.000001)

        res = {}

        res["evidence"] = [out]
        res["uncertainty"] = [uncertainty]
        res["concentration"] = [concentration]
        res["expected_probability"] = [concentration / S]

        return res

    def mc_forward(self, input, num_samples=None):

        with torch.no_grad():
            if num_samples is None:
                num_samples = self.num_samples

            print("Number of Samples = {}".format(num_samples))

            (point_cloud,) = input

            device = point_cloud.device

            for m in self.modules():
                if m.__class__.__name__ == "Dropout":
                    m.train()

            num_batch = torch.unique(point_cloud[:, 0].int()).shape[0]

            pvec = torch.zeros((num_batch, self.num_classes)).to(device)
            logits = torch.zeros((num_batch, self.num_classes)).to(device)
            discrete = torch.zeros((num_batch, self.num_classes)).to(device)

            eye = torch.eye(self.num_classes).int().to(device)

            for i in range(num_samples):
                x = self.encoder(point_cloud)
                out = self.logit_layer(x)
                logits += out
                pred = torch.argmax(out, dim=1)
                pvec += F.softmax(out, dim=1)
                discrete += eye[pred]

            mc_dist = discrete / float(num_samples)
            softmax_probs = pvec / float(num_samples)
            logits = logits / float(num_samples)
            # logits = torch.logit(softmax_probs)
            res = {"softmax": [softmax_probs], "logits": [logits], "mc_dist": [mc_dist]}
            return res

    def standard_forward(self, input, verbose=False):
        print("Forwarding using weight averaging (standard dropout) ...")
        (point_cloud,) = input
        out = self.encoder(point_cloud)
        out = self.logit_layer(out)
        res = {"logits": [out]}
        return res

    def forward(self, input):
        if (not self.training) and (self.mode == "mc_dropout"):
            return self.mc_forward(input)
        elif self.mode == "evidential":
            return self.evidential_forward(input)
        else:
            return self.standard_forward(input)


class MultiParticleTypeLoss(nn.Module):

    def __init__(self, cfg, name="particle_type_loss"):
        super(MultiParticleTypeLoss, self).__init__()

        loss_cfg = cfg.get(name, {})
        self.num_classes = loss_cfg.get("num_classes", 5)
        self.batch_col = loss_cfg.get("batch_col", 0)
        self.target_col = loss_cfg.get("target_col", 9)
        self.balance_classes = loss_cfg.get("balance_classes", False)

        reduction = "mean" if not self.balance_classes else "sum"
        self.xentropy = nn.CrossEntropyLoss(ignore_index=-1, reduction=reduction)

        self.split_input_mode = loss_cfg.get("split_input_as_tg_batch", False)

    def forward_tg(self, out, valid_labels):

        logits = out["logits"][0]
        clusts = out["clusts"][0]

        labels = get_cluster_label(valid_labels, clusts, self.target_col)

        return [logits], [labels]

    def forward(self, out, type_labels):

        valid_labels = type_labels[0][type_labels[0][:, 9] < self.num_classes]

        if self.split_input_mode:
            logits, labels = self.forward_tg(out, valid_labels)

        else:
            logits = out["logits"][0]
            clusts = out["clusts"][0]
            labels = [
                get_cluster_label(
                    valid_labels[valid_labels[:, self.batch_col] == b],
                    clusts[b],
                    self.target_col,
                )
                for b in range(len(clusts))
                if len(clusts[b])
            ]

        if not len(labels):
            res = {
                "loss": torch.tensor(
                    0.0, requires_grad=True, device=valid_labels.device
                ),
                "accuracy": 1.0,
            }
            for c in range(self.num_classes):
                res[f"accuracy_class_{c}"] = 1.0
            return res

        labels = torch.tensor(
            np.concatenate(labels), dtype=torch.long, device=valid_labels.device
        )
        logits = torch.cat(logits, axis=0)

        if not self.balance_classes:
            loss = self.xentropy(logits, labels)
        else:
            classes, counts = labels[labels > -1].unique(return_counts=True)
            weights = torch.sum(counts) / counts / self.num_classes
            loss = 0.0
            for i, c in enumerate(classes):
                class_mask = labels == c
                loss += (
                    weights[i]
                    * self.xentropy(logits[class_mask], labels[class_mask])
                    / torch.sum(counts)
                )

        pred = torch.argmax(logits, dim=1)
        accuracy = float(torch.sum(pred[labels > -1] == labels[labels > -1])) / float(
            labels[labels > -1].shape[0]
        )

        res = {"loss": loss, "accuracy": accuracy}

        for c in range(self.num_classes):
            mask = labels == c
            res[f"accuracy_class_{c}"] = (
                float(torch.sum(pred[mask] == labels[mask])) / float(torch.sum(mask))
                if torch.sum(mask)
                else 1.0
            )

        return res


class MultiLabelCrossEntropy(nn.Module):

    def __init__(self, cfg, name="duq_particle_classifier"):
        super(MultiLabelCrossEntropy, self).__init__()
        self.xentropy = nn.BCELoss(reduction="none")
        self.num_classes = 5
        model_cfg = cfg.get(name, {})
        self.grad_w = model_cfg.get("grad_w", 0.0)
        self.grad_penalty = model_cfg.get("grad_penalty", True)

    @staticmethod
    def calc_gradient_penalty(x, y_pred):
        """
        Code From the DUQ main Github Repository:
        https://github.com/y0ast/deterministic-uncertainty-quantification

        Author: Joost van Amersfoort
        """
        gradients = torch.autograd.grad(
            outputs=y_pred,
            inputs=x,
            grad_outputs=torch.ones_like(y_pred),
            create_graph=True,
        )[0]

        gradients = gradients.flatten(start_dim=1)

        # L2 norm
        grad_norm = gradients.norm(2, dim=1)

        # Two sided penalty
        gradient_penalty = ((grad_norm - 1) ** 2).mean()

        # One sided penalty - down
        #     gradient_penalty = F.relu(grad_norm - 1).mean()

        return gradient_penalty

    def forward(self, out, type_labels):
        probas = out["score"][0]
        device = probas.device
        labels_one_hot = torch.eye(self.num_classes)[type_labels[0][:, 0].long()].to(
            device=device
        )
        loss1 = self.xentropy(probas, labels_one_hot)
        pred = torch.argmax(probas, dim=1)
        labels = type_labels[0][:, 0].long()

        # Comptue gradient penalty
        loss2 = 0
        if self.grad_penalty:
            loss2 = self.calc_gradient_penalty(out["input"][0], probas)

        loss1 = loss1.sum(dim=1).mean()
        loss = loss1 + self.grad_w * loss2

        accuracy = float(torch.sum(pred == labels)) / float(labels.shape[0])

        res = {
            "loss": loss,
            "loss_embedding": float(loss1),
            "loss_grad_penalty": float(loss2),
            "accuracy": accuracy,
        }

        print(res)

        acc_types = {}
        for c in labels.unique():
            mask = labels == c
            acc_types["accuracy_class_{}".format(int(c))] = float(
                torch.sum(pred[mask] == labels[mask])
            ) / float(torch.sum(mask))
        return res


class EvidentialLearningLoss(nn.Module):

    def __init__(self, cfg, name="evidential_learning_loss"):
        super(EvidentialLearningLoss, self).__init__()
        self.loss_config = cfg.get(name, {})
        self.evd_loss_name = self.loss_config.get("evd_loss_name", "edl_sumsq")
        self.num_classes = self.loss_config.get("num_classes", 5)
        self.num_total_iter = self.loss_config.get("num_total_iter", 50000)
        self.loss_fn = EVDLoss(self.evd_loss_name, "mean", T=self.num_total_iter)

    def forward(self, out, type_labels, iteration=0):

        alpha = out["concentration"][0]
        probs = out["expected_probability"][0]
        device = alpha.device

        labels = type_labels[0][:, 0].to(dtype=torch.long)

        labels_onehot = torch.eye(self.num_classes, device=device)[labels]

        loss_batch = self.loss_fn(alpha, labels_onehot, t=iteration)
        loss = loss_batch.mean()
        pred = torch.argmax(probs, dim=1)

        accuracy = float(torch.sum(pred == labels)) / float(labels.shape[0])

        res = {"loss": loss, "accuracy": accuracy}

        acc_types = {}
        for c in labels.unique():
            mask = labels == c
            acc_types["accuracy_class_{}".format(int(c))] = float(
                torch.sum(pred[mask] == labels[mask])
            ) / float(torch.sum(mask))

        return res



"""Visualization tools for confusion matrices."""

import matplotlib as mpl
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt

mpl.rcParams.update({"font.size": 18})
mpl.rcParams.update({"figure.autolayout": True})


def draw_confusion_matrix(
    file_path,
    num_classes=None,
    mapping=None,
    figure_name="confmat",
    show_counts=False,
    class_names=None,
    figsize=(9, 6),
    norm_axis=0,
):
    """Draws the confusion matrix from a file produced by the analysis
    script used to evaluate the classification accuracy.

    Parameters
    ----------
    file_path : str
        Path to the file which contains the classification metrics
    num_classes : int, optional
        Number of classes to represent
    mapping : dict, optional
        Mapping between the stored class and a redefined set of classes
    show_counts : bool, default False
        Show the number of entries in the contingency matrix
    class_names : list, optional
        Labels for each class
    figsize : tuple, default (9, 6)
        Figure size
    norm_axis : int, default 0
        Normalization axis (0: recall, 1: precision)
    """
    # Load the file
    data = pd.read_csv(file_path)

    # Load the contingency matrix
    if "score_0" in data.keys():
        hist = build_matrix(data, num_classes, mapping)
    else:
        hist = rebuild_matrix(data, num_classes, mapping)

    # Initialize figure
    fig = plt.figure(figsize=figsize)
    fig.patch.set_alpha(0)

    # Normalize the histogram counts to the total number of entries in each true class bin
    assert norm_axis in (
        0,
        1,
    ), "The normalization axis must be 0 (recall) or 1 (precision)."
    norms = np.sum(hist, axis=norm_axis)
    if norm_axis == 0:
        hist_norm = hist / norms
    else:
        hist_norm = hist / norms[:, None]

    # Initialize plot, fill
    num_classes = len(hist)
    xedges = yedges = -0.5 + np.arange(0, num_classes + 1)
    plt.pcolormesh(xedges, yedges, hist_norm, cmap="Blues")
    for i in range(num_classes):
        for j in range(num_classes):
            label = (
                "{:0.3f}\n({})".format(hist_norm[i, j], int(hist[i, j]))
                if show_counts
                else "{:0.3f}".format(hist_norm[i, j])
            )
            plt.text(
                j,
                i,
                label,
                color="white" if hist_norm[i, j] > 0.5 else "black",
                ha="center",
                va="center",
            )

    # Set axes style and labels
    plt.xlabel("Class label")
    plt.ylabel("Class prediction")
    if class_names is not None:
        assert (
            len(class_names) == num_classes
        ), "Must provide one class label per class."
        plt.xticks(np.arange(num_classes), labels=class_names)
        plt.yticks(np.arange(num_classes), labels=class_names)
    plt.colorbar()

    # Save and show
    plt.savefig(f"{figure_name}.png", bbox_inches="tight")
    plt.show()


def build_matrix(data, num_classes=None, mapping=None):
    """Builds a confusion matrix from a pixel-wise storage file.

    Parameters
    ----------
    data : pd.Dataframe
        Dataframe which contains the pixel label/predictions
    num_classes : int, optional
        Number of classes to represent
    mapping : dict, optional
        Mapping between the stored class and a redefined set of classes
    """
    # If the number of classes is not specified, fetch from the file
    if num_classes is None:
        if mapping is not None:
            num_classes = len(mapping)
        else:
            classes = []
            for k in data.keys():
                if k.startswith("score"):
                    classes.append(int(k[-1]))
            num_classes = np.max(classes) + 1

    assert (
        mapping is None or len(mapping) == num_classes
    ), "The number of classes should match those in the map."

    # Build the confusion matrix
    hist = np.histogram2d(
        data.pred,
        data.label,
        bins=[num_classes, num_classes],
        range=[[0, num_classes], [0, num_classes]],
    )[0]

    return hist


def rebuild_matrix(data, num_classes=None, mapping=None):
    """Builds a confusion matrix from an entry-wise storage file.

    Parameters
    ----------
    data : pd.Dataframe
        Dataframe which contains the flattened matrix
    num_classes : int, optional
        Number of classes to represent
    mapping : dict, optional
        Mapping between the stored class and a redefined set of classes
    """
    # If the number of classes is not specified, fetch from the file
    if num_classes is None:
        if mapping is not None:
            num_classes = len(mapping)
        else:
            classes = []
            for k in data.keys():
                if k.startswith("count"):
                    classes.append(int(k[-1]))
            num_classes = np.max(classes) + 1

    assert (
        mapping is None or len(mapping) == num_classes
    ), "The number of classes should match those in the map."

    # Rebuild confusion matrix
    hist = np.empty((num_classes, num_classes), dtype=np.int64)
    for i in range(num_classes):
        for j in range(num_classes):
            if mapping is None or not (i in mapping and j in mapping):
                hist[i, j] = np.sum(data[f"count_{i}{j}"])
            else:
                hist[i, j] = 0
                for k in mapping[i]:
                    for l in mapping[j]:
                        hist[i, j] += np.sum(data[f"count_{k}{l}"])

    return hist



"""Analysis script used to evaluate the semantic segmentation accuracy."""

from spine.ana.base import AnaBase
from spine.data.out import RecoInteraction, TruthInteraction

__all__ = ["FlashMatchingAna"]


class FlashMatchingAna(AnaBase):
    """Class which computes and stores the necessary data to build a
    semantic segmentation confusion matrix.
    """

    # Name of the analysis script (as specified in the configuration)
    name = "flash_match_eval"

    # Valid match modes
    _match_modes = ("reco_to_truth", "truth_to_reco", "both", "all")

    # Default object types when a match is not found
    _default_objs = (
        ("reco_interactions", RecoInteraction()),
        ("truth_interactions", TruthInteraction()),
    )

    def __init__(
        self,
        time_window=None,
        neutrino_only=True,
        max_num_flashes=1,
        match_mode="both",
        **kwargs,
    ):
        """Initialize the analysis script.

        Parameters
        ----------
        time_window : List[float], optional
            Time window (in ns) for which interactions must have matched flash
        neutrino_only : bool, default False
            If `True`, only check if neutrino in-time activity is matched for
            the efficiency measurement (as opposed to any in-time activity)
        max_num_flashes : int
            Maximum number of flash matches to store
        match_mode : str, default 'both'
            If reconstructed and truth are available, specified which matching
            direction(s) should be saved to the log file.
        **kwargs : dict, optional
            Additional arguments to pass to :class:`AnaBase`
        """
        # Initialize the parent class
        super().__init__("interaction", "both", **kwargs)

        # Store basic parameters
        self.time_window = time_window
        self.neutrino_only = neutrino_only

        # Store default objects as a dictionary
        self.default_objs = dict(self._default_objs)

        # Store the matching mode
        self.match_mode = match_mode
        assert match_mode in self._match_modes, (
            f"Invalid matching mode: {self.match_mode}. Must be one "
            f"of {self._match_modes}."
        )

        # Make sure the matches are loaded, initialize the output files
        keys = {}
        for prefix in self.prefixes:
            if prefix == "reco" and match_mode != "truth_to_reco":
                keys["interaction_matches_r2t"] = True
                keys["interaction_matches_r2t_overlap"] = True
                self.initialize_writer("reco")
            if prefix == "truth" and match_mode != "reco_to_truth":
                keys["interaction_matches_t2r"] = True
                keys["interaction_matches_t2r_overlap"] = True
                self.initialize_writer("truth")

        self.update_keys(keys)

        # List the interaction attributes to be stored
        nu_attrs = ("energy_init",) if neutrino_only else ()
        flash_attrs = (
            "is_flash_matched",
            "flash_ids",
            "flash_times",
            "flash_scores",
            "flash_total_pe",
            "flash_hypo_pe",
        )
        flash_lengths = {
            k: max_num_flashes for k in ["flash_ids", "flash_times", "flash_scores"]
        }

        self.reco_attrs = ("id", "size", "is_contained", "topology", *flash_attrs)
        self.truth_attrs = (
            "id",
            "size",
            "is_contained",
            "nu_id",
            "t",
            "topology",
            *nu_attrs,
        )

        self.reco_lengths = flash_lengths
        self.truth_lengths = None

    def process(self, data):
        """Store the flash matching metrics for one entry.

        Parameters
        ----------
        data : dict
            Dictionary of data products
        """
        # Loop over the matching directions
        prefixes = {"reco": "truth", "truth": "reco"}
        for source, target in prefixes.items():
            # Loop over the match pairs
            src_attrs = getattr(self, f"{source}_attrs")
            src_lengths = getattr(self, f"{source}_lengths")
            tgt_attrs = getattr(self, f"{target}_attrs")
            tgt_lengths = getattr(self, f"{target}_lengths")
            match_suffix = f"{source[0]}2{target[0]}"
            match_key = f"interaction_matches_{match_suffix}"
            for idx, (obj_i, obj_j) in enumerate(data[match_key]):
                # Check that the source interaction is of interest
                if obj_i.is_truth:
                    # If the source object is a true interaction, check if it
                    # should be matched or not (in time or not)
                    if self.time_window is not None and (
                        obj_i.t < self.time_window[0] or obj_i.t > self.time_window[1]
                    ):
                        continue

                    # If requested, check that the in-time activity is a neutrino
                    if self.neutrino_only and obj_i.nu_id < 0:
                        continue

                else:
                    # If the source object is a reco interaction, check if it
                    # is matched to a flash or not
                    if not obj_i.is_flash_matched:
                        continue

                # Store information about the corresponding reco interaction
                # and the flash associated with it (if any)
                src_dict = obj_i.scalar_dict(src_attrs, src_lengths)
                if obj_j is not None:
                    tgt_dict = obj_j.scalar_dict(tgt_attrs, tgt_lengths)
                else:
                    default_obj = self.default_objs[f"{target}_interactions"]
                    tgt_dict = default_obj.scalar_dict(tgt_attrs, tgt_lengths)

                src_dict = {f"{source}_{k}": v for k, v in src_dict.items()}
                tgt_dict = {f"{target}_{k}": v for k, v in tgt_dict.items()}

                # Get the match quality
                overlap = data[f"{match_key}_overlap"][idx]

                # Build row dictionary and store
                row_dict = {**src_dict, **tgt_dict}
                row_dict.update({"match_overlap": overlap})

                self.append(source, **row_dict)



"""Applies conversion form ADC to number ionization electrons."""

from typing import Dict, List, Optional, Union

import numpy as np

from .constant import CalibrationConstant

__all__ = ["GainCalibrator"]


class GainCalibrator:
    """Converts all charge depositions in ADC to a number of electrons.

    It can either use a flat converstion factor or one per TPC in the detector. It
    can also use a SQLite database to provide gain values which depend on the run.
    """

    name = "gain"

    def __init__(
        self,
        num_tpcs: int,
        gain: Optional[Union[float, List[float]]] = None,
        gain_db: Optional[Union[str, Dict[int, Union[float, List[float]]]]] = None,
    ):
        """Initialize the recombination model and its constants.

        Parameters
        ----------
        num_tpcs : int
            Number of TPCs in the detector
        gain : Union[float, List[float]], optional
            Conversion factor from ADC to electrons (unique or per tpc)
        gain_db : str, optional
            Path to a SQLite db file which maps [run, cryo, tpc] sets
        """
        # Initialize the gain calibration constant
        self.gain = CalibrationConstant(num_tpcs=num_tpcs, value=gain, database=gain_db)

    def process(
        self, values: np.ndarray, tpc_id: int, run_id: Optional[int] = None
    ) -> np.ndarray:
        """Converts deposition values from ADC to a number of electrons.

        Parameters
        ----------
        values : np.ndarray
            (N) array of depositions in ADC in a specific TPC
        tpc_id : int
            ID of the TPC to use

        Returns
        -------
        np.ndarray
            (N) array of depositions in number of electrons
        """
        # Apply the gain factor to all values in the current TPC
        return values * self.gain.get(tpc_id, run_id)



"""GrapPA: Graph Neural Network for Particle Aggregation.

This module implements the GrapPA (Graph Particle Aggregation) architecture,
a graph neural network designed for clustering and grouping particle instances.

GrapPA learns to aggregate fragment-level features into particle-level clusters
through message passing and edge classification, enabling particle instance
segmentation and identification.
"""

import numpy as np
import torch
from torch import nn

from spine.data import TensorBatch
from spine.utils.enums import enum_factory
from spine.utils.globals import GROUP_COL, LOWES_SHP, SHAPE_COL, TRACK_SHP
from spine.utils.gnn.cluster import (
    form_clusters_batch,
    get_cluster_label_batch,
    get_cluster_primary_label_batch,
)
from spine.utils.gnn.evaluation import (
    node_assignment_batch,
    node_assignment_score_batch,
)

from .layer.common.dbscan import DBSCAN
from .layer.factories import final_factory
from .layer.gnn.factories import (
    edge_encoder_factory,
    edge_loss_factory,
    global_encoder_factory,
    global_loss_factory,
    gnn_model_factory,
    graph_factory,
    node_encoder_factory,
    node_loss_factory,
)

__all__ = ["GrapPA", "GrapPALoss"]


class GrapPA(torch.nn.Module):
    """Graph Particle Aggregator (GrapPA) model.

    This class mostly acts as a wrapper that will hand the graph data
    to the underlying graph neural network (GNN).

    When trained standalone, this model must be provided with a cluster
    label tensor, allowing it to build a set of intput clusters based on the
    label boundaries of the clusters and their semantic types.

    Typical configuration can look like this:

    .. code-block:: yaml

        model:
          name: grappa
          modules:
            grappa:
              nodes:
                <dictionary of arguments to specify the input type>
              graph:
                name: <name of the input graph type>
                <dictionary of arguments to specify the graph>
              node_encoder:
                name: <name of the type of node encoder>
                <dictionary of arguments to specify the node encoder>
              edge_encoder:
                name: <name of the type of edge encoder>
                <dictionary of arguments to specify the edge encoder>
              global_encoder:
                name: <name of the type of global encoder>
                <dictionary of arguments to specify the global encoder>
              gnn_model:
                name: <name of the type of backbone GNN feature extractor>
                <dictionary of arguments to specify the GNN>

    See configuration files prefixed with `grappa_` under the `config`
    directory for detailed examples of working configurations.

    See Also
    --------
    :class:`GrapPALoss`
    """

    # TODO: update
    MODULES = [
        ("grappa", ["base", "dbscan", "node_encoder", "edge_encoder", "gnn_model"]),
        "grappa_loss",
    ]

    def __init__(self, grappa, grappa_loss=None):
        """Initialize the GrapPA model.

        Parameters
        ----------
        grappa : dict
            Model configuration
        grappa_loss : dict, optional
            Loss configuration
        """
        # Initialize the parent class
        super().__init__()

        # Process the model configuration
        self.process_model_config(**grappa)

    def process_model_config(
        self,
        nodes,
        graph,
        gnn_model,
        node_encoder,
        edge_encoder=None,
        global_encoder=None,
        dbscan=None,
    ):
        """Process the top-level configuration block.

        This dispatches each block to its own configuration processor.

        Parameters
        ----------
        nodes : dict
            Input node configuration
        graph : dict
            Input graph configuration
        gnn_model : dict
            Underlying graph neural network configuration
        node_encoder : dict
            Node encoder configuration
        edge_encoder : dict
            Edge encoder configuration
        global_encoder : dict, optional
            Global encoder configuration
        dbscan : dict, optional
            DBSCAN fragmentation configuration
        """
        # Store the output types of GNNs
        self.out_types = ["node", "edge", "global"]

        # Process the node configuration
        self.process_node_config(**nodes)

        # Construct the underlying graph neural network
        self.process_gnn_config(**gnn_model)

        # Process the graph configuration
        self.graph_constructor = graph_factory(graph, self.node_type)

        # Process the encoder configurations
        self.node_encoder = node_encoder_factory(node_encoder)

        # Initialize edge encoder
        self.edge_encoder = None
        if edge_encoder is not None:
            self.edge_encoder = edge_encoder_factory(edge_encoder)

        # Initialize the global encoder
        self.global_encoder = None
        if global_encoder is not None:
            self.global_encoder = global_encoder_factory(global_encoder)

        # Process the dbscan fragmenter configuration, if provided
        self.dbscan = None
        if dbscan is not None:
            self.process_dbscan_config(dbscan)

    def process_node_config(
        self,
        source="cluster",
        shapes=None,
        min_size=-1,
        make_groups=False,
        grouping_method="score",
        grouping_through_track=False,
    ):
        """Process the node parameters of the model.

        Parameters
        ----------
        source : str, default 'cluster'
            Column name in the label tensor which contains the input cluster IDs
        shapes : int, optional
            Type of nodes to include in the input. If not specified, include
            all types
        min_size : int, default -1
            Minimum number of voxels in a cluster to be included in the input
        make_groups : bool, default False
            Use edge predictions to build node groups
        grouping_method : str, default 'score'
            Algorithm used to build a node partition
        grouping_through_track : bool, default False
            If `True`, shower objects can only be connected to one track object
        """
        # Parse the node source
        self.node_source = enum_factory("cluster", source)

        # Interpret node type as list of shapes to cluster
        if shapes is None:
            self.node_type = list(np.arange(LOWES_SHP))
        else:
            assert not np.isscalar(
                shapes
            ), "Semantic classes should be provided as a list."
            self.node_type = enum_factory("shape", shapes)

        # Store the node parameters
        self.node_min_size = min_size
        self.make_groups = make_groups
        self.grouping_method = grouping_method
        self.grouping_through_track = grouping_through_track

    def process_gnn_config(
        self, node_pred=None, edge_pred=None, global_pred=None, **gnn_model
    ):
        """Process the GNN backbone structure and the output layers.

        Parameters
        ----------
        node_pred : Union[int, dict], optional
            Number of node predictions. If there are multiple node predictions,
            provide a (key, value) pair for each type of prediction
        edge_pred : Union[int, dict], optional
            Number of edge predictions. If there are multiple edge predictions,
            provide a (key, value) pair for each type of prediction
        global_pred : Union[int, dict], optional
            Number of edge predictions. If there are multiple edge predictions,
            provide a (key, value) pair for each type of prediction
        **gnn_model, dict
            Paramters to initialize the GNN backbone
        """
        # Initialize the GNN backbone
        self.gnn = gnn_model_factory(
            gnn_model,
            node_pred is not None,
            edge_pred is not None,
            global_pred is not None,
        )

        # Initialize output layers based on the configuration
        self.process_final_config(node_pred, "node")
        self.process_final_config(edge_pred, "edge")
        self.process_final_config(global_pred, "global")

    def process_final_config(self, final, prefix):
        """Process a final layer configuration.

        Parameters
        ----------
        final : Union[int, dict]
            Final layer configuration
        prefix : dict
            Name of the final layer
        """
        # If the final layer is not specified, nothing to do here
        if final is None:
            setattr(self, f"{prefix}_pred_keys", [])
            return

        # If the final layer is specified as a number, use linear layer
        if isinstance(final, int):
            final = {"name": "linear", "out_channels": final}

        # Process the configuration dictionary otherwise
        out_keys = []
        in_channels = getattr(self.gnn, f"{prefix}_feature_size")
        if "name" in final:
            # Initialize a single final layer (single prediction of this type)
            out_key = f"{prefix}_pred"
            out_keys.append(out_key)
            setattr(self, out_key, final_factory(in_channels, **final))

        else:
            # Otherwise, initialize one final layer per prediction type
            for key, cfg in final.items():
                # If the final layer is specified as a number, use linear layer
                out_key = f"{prefix}_{key}_pred"
                out_keys.append(out_key)
                if isinstance(cfg, int):
                    cfg = {"name": "linear", "out_channels": cfg}
                setattr(self, out_key, final_factory(in_channels, **cfg))

        setattr(self, f"{prefix}_pred_keys", out_keys)

    def process_dbscan_config(self, shapes=None, min_size=None, **kwargs):
        """Process the DBSCAN fragmenter configuration.

        Parameters
        ----------
        shapes : Union[int, list], optional
            This should not be specified (fetched from the node configuration)
        min_size : Union[int, list], optional
            This should not be specified (fetched from the node configuration)
        **kwargs : dict, optional
            Rest of the DBSCAN configuration
        """
        # Make sure the basic parameters are not specified twice
        assert shapes is not None and min_size is not None, (
            "Do not specify 'shapes' or 'min_size' in the "
            "`dbscan` block, it is shared with the `node` block"
        )

        # Initialize DBSCAN fragmenter
        self.dbscan = DBSCAN(shapes=self.node_type, min_size=self.min_size, **kwargs)

    def forward(
        self,
        data,
        coord_label=None,
        clusts=None,
        shapes=None,
        groups=None,
        points=None,
        extra=None,
    ):
        """Prepares particle clusters and feed them to the GNN model.

        Parameters
        ----------
        data : TensorBatch
            (N, 1 + D + N_f) Tensor of voxel/value pairs
            - N is the the total number of voxels in the image
            - 1 is the batch ID
            - D is the number of dimensions in the input image
            - N_f is 1 (charge/energy) if the clusters (`clusts`) are provided,
              or it needs to contain cluster labels to build them on the fly
        coord_label : TensorBatch, optional
            (P, 1 + 2*D + 2) Tensor of label points (start/end/time/shape)
        clusts : IndexBatch, optional
            (C) List of indexes corresponding to each cluster
        shapes : TensorBatch, optional
            (C) List of cluster semantic class used to define the max length
        groups : TensorBatch, optional
            (C) List of node groups, one per cluster. If specified, will
                remove connections between nodes of a separate group.
        points : TensorBatch, optional
            (C, 3/6) Tensor of start (and end) points
        extra : TensorBatch, optional
            (C, N_f) Batch of features to append to the existing node features

        Returns
        -------
        clusts : IndexBatch
            (C, N_c, N_{c,i}) Cluster indexes
        edge_index : TensorBatch
            (E, 2) Incidence matrix
        node_features : TensorBatch
            (C, N_c,f) Node features
        edge_features : TensorBatch
            (C, N_e,f) Node features
        global_features : TensorBatch
            (C, N_g,f) Global features
        node_pred : TensorBatch
            (C, N_n) Node predictions (logits)
        edge_pred : TensorBatch
            (C, N_e) Edge predictions (logits)
        global_pred : TensorBatch
            (C, N_e) Global predictions (logits)
        """
        # Cast the labels to numpy for the functions run on CPU
        result = {}
        data_np = data.to_numpy()

        # If not provided, form the clusters: a list of list of voxel indices,
        # one list per cluster matching the list of requested class
        if clusts is None:
            if self.dbscan is not None:
                # Use the DBSCAN fragmenter to build the clusters
                seg_label = TensorBatch(data.tensor[:, SHAPE_COL], data.counts)
                clusts, _ = self.dbscan(data, seg_label, coord_label)
            else:
                # Use the label tensor to build the clusters
                clusts = form_clusters_batch(
                    data_np, self.node_min_size, self.node_source, shapes=self.node_type
                )

        result["clusts"] = clusts

        # If the graph edge length cut is class-specific, get the class labels
        if shapes is None and hasattr(self.graph_constructor.max_length, "__len__"):
            if self.node_source == GROUP_COL:
                # For groups, use primary shape to handle Michel/Delta properly
                shapes = get_cluster_primary_label_batch(data_np, clusts, SHAPE_COL)
            else:
                # Just use the shape of the cluster itself otherwise
                shapes = get_cluster_label_batch(data_np, clusts, SHAPE_COL)

            shapes.data = shapes.data.astype(np.int64)

        # Initialize the input graph
        edge_index, _, closest_index = self.graph_constructor(
            data_np, clusts, shapes, groups
        )

        result["edge_index"] = edge_index

        # Fetch the node features
        node_features = self.node_encoder(
            data, clusts, coord_label=coord_label, points=points, extra=extra
        )

        if isinstance(node_features, tuple):
            # If the output of the node encoder is a tuple, separate points
            node_features, points = node_features
            start_points, end_points = points.tensor.split(3, dim=1)

            result["start_points"] = TensorBatch(
                start_points, points.counts, coord_cols=np.array([0, 1, 2])
            )
            result["end_points"] = TensorBatch(
                end_points, points.counts, coord_cols=np.array([0, 1, 2])
            )

        # Fetch the edge features
        edge_features = None
        if self.edge_encoder is not None:
            edge_features = self.edge_encoder(
                data, clusts, edge_index, closest_index=closest_index
            )

        # Feath the global_features
        global_features = None
        if self.global_encoder is not None:
            global_features = self.global_encoder(data, clusts)

        # Bring edge_index and batch_ids to device
        # TODO: try to keep everything (apart from clusts?) on GPU?
        index = torch.tensor(edge_index.index, device=data.tensor.device)
        xbatch = torch.tensor(clusts.batch_ids, device=data.tensor.device)

        # Pass through the model, update results
        out = self.gnn(node_features, index, edge_features, global_features, xbatch)

        # Loop over the necessary node/edge/global predictions, store output
        for t in self.out_types:
            for key in getattr(self, f"{t}_pred_keys"):
                result[key] = getattr(self, key)(out[f"{t}_features"])

        # If requested, build node groups from edge predictions
        if self.make_groups:
            edge_pred_keys = [
                key for key in result if key.startswith("edge") and key.endswith("pred")
            ]
            assert (
                len(edge_pred_keys) > 0
            ), "Must provide edge predictions to build node groups."
            for key in edge_pred_keys:
                edge_pred = result[key].to_numpy()
                prefix = "group" + key.replace("edge", "").replace("_pred", "")
                if self.grouping_method == "threshold":
                    result[f"{prefix}_pred"] = node_assignment_batch(
                        edge_index, edge_pred, clusts
                    )
                elif self.grouping_method == "score":
                    if not self.grouping_through_track:
                        result[f"{prefix}_pred"] = node_assignment_score_batch(
                            edge_index, edge_pred, clusts
                        )
                    else:
                        assert (
                            shapes is not None
                        ), "Must provide shapes to restrict track association."
                        track_node = TensorBatch(
                            shapes.data == TRACK_SHP, counts=shapes.counts
                        )
                        result[f"{prefix}_pred"] = node_assignment_score_batch(
                            edge_index, edge_pred, clusts, track_node
                        )

                else:
                    raise ValueError(
                        "Group prediction algorithm not recognized:",
                        self.grouping_method,
                    )

        return result


class GrapPALoss(torch.nn.modules.loss._Loss):
    """Takes the output of the GrapPA and computes the total loss.

    For use in config:

    ..  code-block:: yaml

        model:
          name: grappa
          modules:
            grappa_loss:
              node_loss:
                name: <name of the node loss>
                <dictionary of arguments to pass to the loss>
              edge_loss:
                name: <name of the edge loss>
                <dictionary of arguments to pass to the loss>
              global_loss:
                name: <name of the global loss>
                <dictionary of arguments to pass to the loss>

    Each of the specific loss blocks can also contain multiple losses by
    providing a name key in a loss block nested below it. Each loss name of a
    specific type should be provided with a corresponding output from GRaPA.

    See configuration files prefixed with `grappa_` under the `config`
    directory for detailed examples of working configurations.
    """

    def __init__(self, grappa_loss, grappa=None):
        """Initialize the GrapPA loss function.

        Parameters
        ----------
        grappa_loss : dict
            Loss configuration
        grappa : dict, optional
            Model configuration
        """
        # Initialize the parent class
        super().__init__()

        # Process the loss configuration
        self.process_loss_config(**grappa_loss)

    def process_loss_config(self, node_loss=None, edge_loss=None, global_loss=None):
        """Process the loss configuration.

        Parameters
        ----------
        node_loss : Union[dict, Dict[dict]], optional
            Node loss configuration
        edge_loss : Union[dict, Dict[dict]], optional
            Edge loss configuration
        global_loss : Union[dict, Dict[dict]], optional
            Global loss configuration
        """
        # Check that there is at least one loss to apply
        self.out_types = ["node", "edge", "global"]
        assert (
            node_loss is not None or edge_loss is not None or global_lsos is not None
        ), (
            "Must provide either a `node_loss`, `edge_loss` or "
            "`global_loss` to the GrapPA loss function."
        )

        # Initialize the node/edge/global losses
        self.process_single_loss_config("node", node_loss, node_loss_factory)
        self.process_single_loss_config("edge", edge_loss, edge_loss_factory)
        self.process_single_loss_config("global", global_loss, global_loss_factory)

    def process_single_loss_config(self, prefix, loss, constructor):
        """Process a loss configuration.

        Parameters
        ----------
        prefix : dict
            Name of the output type to apply the loss to
        loss : Union[int, dict]
            Loss configuration
        constructor : object
            Loss constructor function
        """
        # If the loss is not specified, nothing to do here
        if loss is None:
            setattr(self, f"{prefix}_loss_keys", [])
            return

        # Process the configuration dictionary otherwise
        loss_keys = []
        if "name" in loss:
            # Initialize a single loss
            loss_key = f"{prefix}_loss"
            loss_keys.append(loss_key)
            setattr(self, loss_key, constructor(loss))

        else:
            # Otherwise, initialzie one loss per prediction type
            for key, cfg in loss.items():
                loss_key = f"{prefix}_{key}_loss"
                loss_keys.append(loss_key)
                setattr(self, loss_key, constructor(cfg))

        setattr(self, f"{prefix}_loss_keys", loss_keys)

    def forward(
        self, clust_label, coord_label=None, graph_label=None, iteration=None, **output
    ):
        """Apply the node/edge/global losses to the logits from GrapPA.

        Parameters
        ----------
        clust_label : TensorBatch
            (N, 1 + D + N_f) Tensor of voxel/value pairs
            - N is the the total number of voxels in the image
            - 1 is the batch ID
            - D is the number of dimensions in the input image
            - N_f is is the number of cluster labels
        coord_label : TensorBatch, optional
            (P, 1 + D + 8) Tensor of start/end point labels for each
            true particle in the image
        graph_label : EdgeIndexTensor, optional
            (2, E) Tensor of edges that correspond to physical
            connections between true particle in the image
        iteration : int, optional
            Iteration index
        **output : dict
            Output of the GrapPA model
        """
        # Loop and apply the losses
        result = {}
        num_losses = 0
        loss, accuracy = 0.0, 0.0
        for t in self.out_types:
            loss_keys = getattr(self, f"{t}_loss_keys")
            for key in loss_keys:
                # If the number of loss keys is > 1 for this type of
                # prediction, must rename the prediction appropriately
                extra = {}
                if len(loss_keys) > 1:
                    extra[f"{t}_pred"] = output[key.replace("loss", "pred")]

                # Compute the loss
                out = getattr(self, key)(
                    clust_label=clust_label,
                    coord_label=coord_label,
                    graph_label=graph_label,
                    iteration=iteration,
                    **output,
                    **extra,
                )

                # Increment the loss and accuracy
                loss += out["loss"]
                accuracy += out["accuracy"]
                num_losses += 1

                # Update the result dictionary
                prefix = "_".join(key.split("_")[:-1])
                for k, v in out.items():
                    result[f"{prefix}_{k}"] = v

        # Append the total loss and total accuracy
        result["loss"] = torch.sum(loss) / num_losses
        result["accuracy"] = np.sum(accuracy) / num_losses

        return result



"""k Nearest-neighbor (kNN) graph constructor for GNNs."""

import numba as nb
import numpy as np

from spine.math.linalg import submatrix

from .base import GraphBase

__all__ = ["KNNGraph"]


class KNNGraph(GraphBase):
    """Generates graphs based on the k nearest-neighbor (kNN) graph of the
    input node locations.

    Makes an edge for each nearest neighbor connection.

    See :class:`GraphBase` for attributes/methods shared
    across all graph constructors.
    """

    # Name of the graph constructor (as specified in the configuration)
    name = "knn"

    def __init__(self, k, **kwargs):
        """Initialize the graph constructor.

        This adds the possibility to set the `k` parameter of the kNN graph.

        Parameters
        ----------
        k : int
            Maximum number of nodes a node can be connected to
        **kwargs : dict, optional
            Additional parameters to pass to the :class:`GraphBase`
            constructor.
        """
        # Initialize base class
        super().__init__(**kwargs)

        # Store attribute
        self.k = k

    def generate(self, clusts, dist_mat, **kwargs):
        """Generates an incidence matrix that connects nodes that share an
        edge in their corresponding kNN graph.

        Parameters
        ----------
        clusts : IndexBatch
            (C) Cluster indexes
        dist_mat : Union[np.ndarray, torch.Tensor]
            (C, C) Matrix of pair-wise distances between clusters in the batch
        **kwargs : dict, optional
            Unused graph generation arguments

        Returns
        -------
        np.ndarray
            (2, E) Tensor of edges
        """
        return self._generate(clusts.batch_ids, self.k, dist_mat, self.directed)

    @staticmethod
    @nb.njit(cache=True)
    def _generate(
        batch_ids: nb.int64[:],
        k: nb.int64,
        dist_mat: nb.float64[:, :],
        directed: bool = False,
    ) -> nb.int64[:, :]:
        # Use the available distance matrix to build a kNN graph
        ret = np.empty((0, 2), dtype=np.int64)
        for b in np.unique(batch_ids):
            clust_ids = np.where(batch_ids == b)[0]
            if len(clust_ids) > 1:
                subk = min(k + 1, len(clust_ids))
                submat = submatrix(dist_mat, clust_ids, clust_ids)
                for i in range(len(submat)):
                    idxs = np.argsort(submat[i])[1:subk]
                    edges = np.empty((subk - 1, 2), dtype=np.int64)
                    for j, idx in enumerate(np.sort(idxs)):
                        edges[j] = [clust_ids[i], clust_ids[idx]]
                    if len(edges):
                        ret = np.vstack((ret, edges))

        return ret.T



import MinkowskiEngine as ME
import MinkowskiFunctional as MF
import numpy as np
import torch
import torch.nn as nn

from .act_norm import act_factory
from .blocks import *
from .configuration import setup_cnn_configuration


class UResNeXt(torch.nn.Module):
    """
    UNet Type encoder-decoder network, with atrous convolutions and
    resnext-type blocks.
    """

    def __init__(self, cfg, name="uresnext"):
        super(UResNeXt, self).__init__()
        setup_cnn_configuration(self, cfg, name)

        self.model_cfg = cfg["modules"][name]

        # Configurations
        self.reps = self.model_cfg.get("reps", 2)
        self.depth = self.model_cfg.get("depth", 5)
        self.num_filters = self.model_cfg.get("num_filters", 32)
        self.cardinality = self.model_cfg.get("cardinality", 8)
        self.dilations = [1, 1, 1, 1, 2, 2, 4, 4]
        assert self.num_filters % self.cardinality == 0
        self.nPlanes = [i * self.num_filters for i in range(1, self.depth + 1)]
        # self.nPlanes = [(2**i) * self.num_filters for i in range(self.depth)]
        self.input_kernel = self.model_cfg.get("input_kernel", 3)

        activation = self.activation_name
        activation_args = self.activation_args

        # Initialize Input Layer
        self.input_layer = ME.MinkowskiConvolution(
            self.num_input,
            self.num_filters,
            kernel_size=self.input_kernel,
            stride=1,
            dimension=self.D,
        )

        # Initialize Encoder
        print(self.nPlanes)
        self.encoding_conv = []
        self.encoding_block = []
        for i, F in enumerate(self.nPlanes):
            m = []
            for _ in range(self.reps):
                m.append(
                    ResNeXtBlock(
                        F,
                        F,
                        dimension=self.D,
                        cardinality=self.cardinality,
                        dilations=self.dilations,
                        activation=self.activation_name,
                        activation_args=self.activation_args,
                    )
                )
            m = nn.Sequential(*m)
            self.encoding_block.append(m)
            m = []
            if i < self.depth - 1:
                m.append(ME.MinkowskiBatchNorm(F))
                m.append(act_factory(activation, **activation_args))
                m.append(
                    ME.MinkowskiConvolution(
                        in_channels=self.nPlanes[i],
                        out_channels=self.nPlanes[i + 1],
                        kernel_size=2,
                        stride=2,
                        dimension=self.D,
                    )
                )
            m = nn.Sequential(*m)
            self.encoding_conv.append(m)
        self.encoding_conv = nn.Sequential(*self.encoding_conv)
        self.encoding_block = nn.Sequential(*self.encoding_block)

        # Initialize Decoder
        self.decoding_block = []
        self.decoding_conv = []
        for i in range(self.depth - 2, -1, -1):
            m = []
            m.append(ME.MinkowskiBatchNorm(self.nPlanes[i + 1]))
            m.append(act_factory(activation, **activation_args))
            m.append(
                ME.MinkowskiConvolutionTranspose(
                    in_channels=self.nPlanes[i + 1],
                    out_channels=self.nPlanes[i],
                    kernel_size=2,
                    stride=2,
                    dimension=self.D,
                )
            )
            m = nn.Sequential(*m)
            self.decoding_conv.append(m)
            m = []
            for j in range(self.reps):
                m.append(
                    ResNeXtBlock(
                        self.nPlanes[i] * (2 if j == 0 else 1),
                        self.nPlanes[i],
                        dimension=self.D,
                        cardinality=self.cardinality,
                        dilations=self.dilations,
                        activation=self.activation_name,
                        activation_args=self.activation_args,
                    )
                )
            m = nn.Sequential(*m)
            self.decoding_block.append(m)
        self.decoding_block = nn.Sequential(*self.decoding_block)
        self.decoding_conv = nn.Sequential(*self.decoding_conv)

        # print('Total Number of Trainable Parameters = {}'.format(
        #     sum(p.numel() for p in self.parameters() if p.requires_grad)))

    def encoder(self, x):
        """
        UResNeXt Encoder.

        INPUTS:
            - x (SparseTensor): MinkowskiEngine SparseTensor

        RETURNS:
            - result (dict): dictionary of encoder output with
            intermediate feature planes:
              1) encoderTensors (list): list of intermediate SparseTensors
              2) finalTensor (SparseTensor): feature tensor at
              deepest layer.
        """
        x = self.input_layer(x)
        encoderTensors = [x]
        for i, layer in enumerate(self.encoding_block):
            x = self.encoding_block[i](x)
            encoderTensors.append(x)
            x = self.encoding_conv[i](x)

        result = {"encoderTensors": encoderTensors, "finalTensor": x}
        return result

    def decoder(self, final, encoderTensors):
        """
        UResNeXt Decoder

        INPUTS:
            - encoderTensors (list of SparseTensor): output of encoder.
        RETURNS:
            - decoderTensors (list of SparseTensor):
            list of feature tensors in decoding path at each spatial resolution.
        """
        decoderTensors = []
        x = final
        for i, layer in enumerate(self.decoding_conv):
            eTensor = encoderTensors[-i - 2]
            x = layer(x)
            x = ME.cat((eTensor, x))
            x = self.decoding_block[i](x)
            decoderTensors.append(x)
        return decoderTensors

    def forward(self, input):
        coords = input[:, 0 : self.D + 1].int()
        features = input[:, self.D + 1 :]

        x = ME.SparseTensor(features, coordinates=coords)
        encoderOutput = self.encoder(x)
        encoderTensors = encoderOutput["encoderTensors"]
        finalTensor = encoderOutput["finalTensor"]
        decoderTensors = self.decoder(finalTensor, encoderTensors)

        res = {
            "encoderTensors": encoderTensors,
            "decoderTensors": decoderTensors,
            "finalTensor": finalTensor,
        }
        return res



"""Simple module which define logging module style and returns it."""

import logging
import sys
import warnings

# Configure the formatting of the logger
logging.basicConfig(format="%(message)s", stream=sys.stdout)
# logging.basicConfig(format='[%(levelname)s] %(message)s')
# logging.basicConfig(format='[%(levelname)s][%(name)s] %(message)s')

# Capture warning messages and redirect them through the logger
logging.captureWarnings(True)

# Initialize logger
logger = logging.getLogger("spine")

# Configure the warnings package to only issue warnings once
warnings.simplefilter("ignore")

# Suppress MinkowskiEngine internal problems (nothing we can do about it)
warnings.filterwarnings(
    "ignore", category=DeprecationWarning, message="<class 'Minkowski"
)



"""Modules used to evaluate model performance."""

import torch

__all__ = ["IoUScore"]


class IoUScore(torch.nn.Module):
    """Intersection over union score for binary predictions."""

    name = "iou"

    def forward(self, y_true, y_pred):
        """Evaluate the IoU score for a batch of label and predictions.

        Parameters
        ----------
        y_true : torch.Tensor
            Set of labels
        y_pred : torch.Tensor
            Set of predictions

        Returns
        -------
        float
            IoU score
        """
        # Compute and return
        with torch.no_grad():
            union = (y_true.long() == 1) | (y_pred.long() == 1)
            if not union.any():
                return 0.0

            else:
                intersection = (y_true.long() == 1) & (y_pred.long() == 1)
                return float(intersection.sum() / union.sum())



"""Tools to draw true particles information."""

import numpy as np

from spine.data import Particle
from spine.utils.globals import PART_COL

from .layout import HIGH_CONTRAST_COLORS
from .point import scatter_points

__all__ = ["scatter_particles"]


def scatter_particles(
    cluster_label, particles, part_col=PART_COL, markersize=1, **kwargs
):
    """Builds a graph of true particles in the image.

    Function which returns a graph object per true particle in the
    particle list, provided that the particle deposited energy in the
    detector which appears in the cluster_label tensor.

    Parameters
    ----------
    cluster_label : np.ndarray
        (N, M) Tensor of pixel coordinates and their associated cluster ID
    particles : List[Particle]
        (P) List of true particle objects
    part_col : int
        Index of the column in the label tensor that contains the particle ID
    **kwargs : dict, optional
        List of additional arguments to pass to plotly.graph_objs.Scatter3D that
        make up the output list

    Returns
    -------
    List[plotly.graph_objs.Scatter3D]
        List of particle traces
    """
    # Initialize one graph per particle
    traces = []
    colors = HIGH_CONTRAST_COLORS
    for i, p in enumerate(particles):
        # Get a mask that corresponds to the particle entry, skip if empty
        index = np.where(cluster_label[:, part_col] == i)[0]
        if not index.shape[0]:
            continue

        # If needed, cast the particle labels to the local class
        if not isinstance(p, Particle):
            p = Particle.from_larcv(p)

        # Initialize the information string
        label = f"Particle {p.id}"
        hovertext_dict = {
            "Particle ID": p.id,
            "Group ID": p.group_id,
            "\u03a1arent ID": p.parent_id,
            "Inter. ID": p.interaction_id,
            "Neutrino ID": p.nu_id,
            "Type ID": p.pid,
            "Group primary": p.group_primary,
            "Inter. primary": p.interaction_primary,
            "Shape ID": p.shape,
            "PDG": p.pdg_code,
            "\u03a1arent PDG": p.parent_pdg_code,
            "Anc. PDG": p.ancestor_pdg_code,
            "Process": p.creation_process,
            "\u03a1arent process": p.parent_creation_process,
            "Anc. process": p.ancestor_creation_process,
            "Initial E": f"{p.energy_init:0.1f} MeV",
            "Deposited E": f"{p.energy_deposit:0.1f} MeV",
            "Time": f"{p.t:0.1f} ns",
            "First step": p.first_step,
            "Last step": p.last_step,
            "Position": p.position,
            "Anc. position": p.ancestor_position,
        }

        hovertext = "".join([f"{l:15}: {v}<br>" for l, v in hovertext_dict.items()])

        # Append a scatter plot trace
        trace = scatter_points(
            cluster_label[index],
            color=str(colors[i % len(colors)]),
            hovertext=hovertext,
            markersize=markersize,
            name=label,
            **kwargs,
        )

        traces += trace

    return traces



"""Module with a base class for all batched data structures."""

from dataclasses import dataclass
from typing import Union

import numpy as np

from spine.utils.conditional import torch


@dataclass(eq=False)
class BatchBase:
    """Base class for all types of batched data.

    Attributes
    ----------
    data : Union[list, np.ndarray, torch.Tensor]
        Batched data
    counts : Union[np.ndarray, torch.Tensor]
        (B) Number of data elements in each entry of the batch
    edges : Union[np.ndarray, torch.Tensor]
        (B+1) Edges separating the entries in the batch
    batch_size : int
        Number of entries that make up the batched data
    """

    data: Union[np.ndarray, torch.Tensor]
    counts: Union[np.ndarray, torch.Tensor]
    edges: Union[np.ndarray, torch.Tensor]
    batch_size: int

    def __init__(self, data, is_sparse=False, is_list=False):
        """Shared initializations across all types of batched data.

        Parameters
        ----------
        data : Union[np.ndarray, torch.Tensor]
            Batched data
        is_sparse : bool, default False
            If initializing from an ME sparse data, flip to True
        is_list : bool, default False
            Whether the underlying data is a list of tensors
        """
        # Store the datatype
        self.is_numpy = not is_sparse and not isinstance(data, torch.Tensor)
        self.is_sparse = is_sparse
        self.is_list = is_list

        # Store the datatype
        self.dtype = data.dtype

        # Store the device
        self.device = None
        if not self.is_numpy:
            ref = data if not is_sparse else data.F
            self.device = ref.device

    def __len__(self):
        """Returns the number of entries that make up the batch."""
        return self.batch_size

    def __eq__(self, other):
        """Checks that all attributes of two class instances are the same.

        This overloads the default dataclass `__eq__` method to include an
        appopriate check for vector (numpy) attributes.

        Parameters
        ----------
        other : obj
            Other instance of the same object class

        Returns
        -------
        bool
            `True` if all attributes of both objects are identical
        """
        # Check that the two objects belong to the same class
        if self.__class__ != other.__class__:
            return False

        # Check that all attributes are identical
        for k, v in self.__dict__.items():
            v_other = getattr(other, k)
            if v is None:
                # If not filled, make sure neither are
                if v_other is not None:
                    return False

            elif np.isscalar(v) or isinstance(v, np.dtype):
                # For scalars, regular comparison will do
                if v_other != v:
                    return False

            else:
                # For vectors, compare all elements
                v_other = getattr(other, k)
                if v.shape != v_other.shape or (v_other != v).any():
                    return False

        return True

    @property
    def shape(self):
        """Shape of the underlying data.

        Returns
        -------
        tuple
            Tuple of sizes in each dimension
        """
        if not self.is_list:
            return self.data.shape
        else:
            return (len(self.data),)

    @property
    def splits(self):
        """Boundaries needed to split the data into its constituents.

        Returns
        -------
        Union[np.ndarray, torch.Tensor]
            (B-1) One split per batch boundary
        """
        return self.edges[1:-1]

    def get_counts(self, batch_ids, batch_size):
        """Finds the number of elements in each entry, provided a batch ID list.

        Parameters
        ----------
        batch_ids : Union[np.ndarray, torch.Tensor]
            List of batch IDs
        batch_size : int
            Number of entries that make up the batched data

        Returns
        -------
        np.ndarray
            (B) Length of each entry
        """
        # Get the count list
        device = None if self.is_numpy else batch_ids.device
        counts = self._zeros(batch_size, device)
        if len(batch_ids):
            # Find the length of each batch ID in the input index
            uni, cnts = self._unique(batch_ids)
            counts[self._as_long(uni)] = cnts

        return counts

    def get_edges(self, counts):
        """Finds the edges between successive entries in the batch.

        Parameters
        ----------
        counts : Union[np.ndarray, torch.Tensor]
            (B)Length of each entry

        Returns
        -------
        np.ndarray
            (B+1) Edges of successive entries in the batch
        """
        # Get the edge list
        device = None if self.is_numpy else counts.device
        edges = self._zeros(len(counts) + 1, device)
        cumsum = self._cumsum(counts)
        edges[1:] = cumsum

        return edges

    def _empty(self, x):
        if self.is_numpy:
            return np.empty(x, dtype=np.int64)
        else:
            return torch.empty(x, dtype=torch.long, device=self.device)

    def _zeros(self, x, device=None):
        if self.is_numpy:
            return np.zeros(x, dtype=np.int64)
        else:
            return torch.zeros(x, dtype=torch.long, device=device)

    def _ones(self, x):
        if self.is_numpy:
            return np.ones(x, dtype=np.int64)
        else:
            return torch.ones(x, dtype=torch.long, device=self.device)

    def _as_long(self, x):
        if self.is_numpy:
            return np.asarray(x, dtype=np.int64)
        else:
            # Always on CPU. This is because splits are supposed to be on
            # CPU regardless of the location of the underlying data
            return torch.as_tensor(x, dtype=torch.long, device="cpu")

    def _unique(self, x):
        if self.is_numpy:
            return np.unique(x, return_counts=True)
        else:
            return torch.unique(x, return_counts=True)

    def _transpose(self, x):
        if self.is_numpy:
            return np.transpose(x)
        else:
            return torch.transpose(x, 0, 1)

    def _sum(self, x):
        if self.is_numpy:
            return np.sum(x)
        else:
            return torch.sum(x)

    def _cumsum(self, x):
        if self.is_numpy:
            return np.cumsum(x)
        else:
            return torch.cumsum(x, dim=0)

    def _arange(self, x):
        if self.is_numpy:
            return np.arange(x)
        else:
            return torch.arange(x, device=self.device)

    def _cat(self, x):
        if self.is_numpy:
            return np.concatenate(x)
        else:
            return torch.cat(x, dim=0)

    def _split(self, *x):
        if self.is_list:
            return np.split(*x)
        else:
            return np.split(*x) if self.is_numpy else torch.tensor_split(*x)

    def _stack(self, x):
        return np.vstack(x) if self.is_numpy else torch.stack(x)

    def _repeat(self, *x):
        return np.repeat(*x) if self.is_numpy else torch.repeat_interleave(*x)

    def _to_numpy(self, x):
        return x.cpu().detach().numpy()

    def _to_tensor(self, x, dtype=None, device=None):
        return torch.as_tensor(x, dtype=dtype, device=device)



"""Module with a dataclass targeted at a batch index or list of indexes."""

from dataclasses import dataclass
from typing import Union
from warnings import warn

import numpy as np

from spine.utils.conditional import torch
from spine.utils.docstring import inherit_docstring

from .base import BatchBase

__all__ = ["IndexBatch"]


@dataclass(eq=False)
@inherit_docstring(BatchBase)
class IndexBatch(BatchBase):
    """Batched index with the necessary methods to slice it.

    Attributes
    ----------
    offsets : Union[np.ndarray, torch.Tensor]
        (B) Offsets between successive indexes in the batch
    single_counts : Union[np.ndarray, torch.Tensor]
        (I) Number of index elements per index in the index list. This
        is the same as counts if the underlying data is a single index
    """

    offsets: Union[np.ndarray, torch.Tensor]
    single_counts: Union[np.ndarray, torch.Tensor]

    def __init__(
        self,
        data,
        offsets,
        counts=None,
        single_counts=None,
        batch_ids=None,
        batch_size=None,
        default=None,
        is_numpy=True,
    ):  # TODO is_numpy does nothing
        """Initialize the attributes of the class.

        Parameters
        ----------
        data : Union[np.ndarray, torch.Tensor,
                     List[Union[np.ndarray, torch.Tensor]]]
            Simple batched index or list of indexes
        offsets : Union[List[int], np.ndarray, torch.Tensor]
            (B) Offsets between successive indexes in the batch
        counts : Union[List[int], np.ndarray, torch.Tensor], optional
            (B) Number of indexes in the batch
        single_counts : Union[List[int], np.ndarray, torch.Tensor], optional
            (I) Number of index elements per index in the index list. This
            is the same as counts if the underlying data is a single index
        batch_ids : Union[List[int], np.ndarray, torch.Tensor], optional
            (I) Batch index of each of the clusters. If not specified, the
            assumption is that each count corresponds to a specific entry
        batch_size : int, optional
            Number of entries in the batch. Must be specified along batch_ids
        is_numpy : bool, default True
            Default type of index. Provide if `data` may be empty
        """
        # Check weather the input is a single index or a list
        is_list = isinstance(data, (list, tuple)) or data.dtype == object

        # Initialize the base class
        if not is_list:
            init_data = data

        elif len(data):
            init_data = data[0]

        else:
            if default is None:
                warn(
                    "The input index data is an empty list without a default "
                    "index. Will use numpy as an underlying representation."
                )
                default = np.empty(0, dtype=np.int64)

            init_data = default

        super().__init__(init_data, is_list=is_list)

        # Get the counts if they are not provided for free
        if counts is None:
            assert (
                batch_ids is not None and batch_size is not None
            ), "Must provide `batch_size` alongside `batch_ids`."
            counts = self.get_counts(batch_ids, batch_size)

        else:
            batch_size = len(counts)

        # Get the number of index elements per entry in the batch
        if single_counts is None:
            assert (
                not self.is_list
            ), "When initializing an index list, provide `single_counts`."
            single_counts = counts
        else:
            assert len(single_counts) == len(
                data
            ), "There must be one single count per index in the list."

        # Cast
        counts = self._as_long(counts)
        single_counts = self._as_long(single_counts)
        offsets = self._as_long(offsets)

        # Do a couple of basic sanity checks
        assert self._sum(counts) == len(
            data
        ), "The `counts` provided must add up to the index length."
        assert len(counts) == len(
            offsets
        ), "The number of `offsets` must match the number of `counts`."

        # Get the boundaries between successive index using the counts
        edges = self.get_edges(counts)

        # Store the attributes
        self.data = data
        self.counts = counts
        self.single_counts = single_counts
        self.edges = edges
        self.offsets = offsets
        self.batch_size = batch_size

    def __getitem__(self, batch_id):
        """Returns a subset of the index corresponding to one entry.

        Parameters
        ----------
        batch_id : int
            Entry index
        """
        # Make sure the batch_id is sensible
        if batch_id >= self.batch_size:
            raise IndexError(
                f"Index {batch_id} out of bound for a batch size "
                f"of ({self.batch_size})"
            )

        # Return
        lower, upper = self.edges[batch_id], self.edges[batch_id + 1]
        if not self.is_list:
            return self.data[lower:upper] - self.offsets[batch_id]

        else:
            entry = np.empty(upper - lower, dtype=object)
            for i, index in enumerate(self.data[lower:upper]):
                entry[i] = index - self.offsets[batch_id]

            return entry

    @property
    def index(self):
        """Alias for the underlying data stored.

        Returns
        -------
        Union[np.ndarray, torch.Tensor]
            Underlying index
        """
        assert (
            not self.is_list
        ), "Underlying data is not a single index, use `index_list`"

        return self.data

    @property
    def index_list(self):
        """Alias for the underlying data list stored.

        Returns
        -------
        List[Union[np.ndarray, torch.Tensor]]
            Underlying index list
        """
        assert self.is_list, "Underlying data is a single index, use `index`"

        return self.data

    @property
    def full_index(self):
        """Returns the index combining all sub-indexes, if relevant.

        Returns
        -------
        Union[np.ndarray, torch.Tensor]
            (N) Complete concatenated index
        """
        if not self.is_list:
            return self.data
        else:
            return self._cat(self.data) if len(self.data) else self._empty(0)

    @property
    def index_ids(self):
        """Returns the ID of the index in the list each element belongs to.

        Returns
        -------
        Union[np.ndarray, torch.Tensor]
            (M) List of index IDs for each element
        """
        assert self.is_list, "Underlying data must be a list of index"

        return self._repeat(self._arange(len(self.data)), self.single_counts)

    @property
    def full_counts(self):
        """Returns the total number of elements in each batch entry.

        Returns
        -------
        Union[np.ndarray, torch.Tensor]
            (B) Number of elements in each batch entry
        """
        if not self.is_list:
            return self.counts
        else:
            full_counts = self._empty(self.batch_size)
            for b in range(self.batch_size):
                lower, upper = self.edges[b], self.edges[b + 1]
                full_counts[b] = self._sum(self.single_counts[lower:upper])

            return self._as_long(full_counts)

    @property
    def batch_ids(self):
        """Returns the batch ID of each index in the list.

        Returns
        -------
        Union[np.ndarray, torch.Tensor]
            (I) Batch ID array, one per index in the list
        """
        return self._repeat(self._arange(self.batch_size), self.counts)

    @property
    def full_batch_ids(self):
        """Returns the batch ID of each element in the full index list.

        Returns
        -------
        Union[np.ndarray, torch.Tensor]
            (N) Complete batch ID array, one per element
        """
        return self._repeat(self._arange(self.batch_size), self.full_counts)

    def split(self):
        """Breaks up the index batch into its constituents.

        Returns
        -------
        List[List[Union[np.ndarray, torch.Tensor]]]
            List of list of indexes per entry in the batch
        """
        # Cast to numpy object array to be able to use split
        if self.is_list and not isinstance(self.data, np.ndarray):
            data_np = np.empty(len(self.data), dtype=object)
            data_np[:] = self.data
        else:
            data_np = self.data

        # Split, offset
        indexes = np.split(data_np, self.splits)
        for batch_id in range(self.batch_size):
            indexes[batch_id] = indexes[batch_id] - self.offsets[batch_id]

        return indexes

    def merge(self, index_batch):
        """Merge this index batch with another.

        Parameters
        ----------
        index_batch : IndexBatch
            Other index batch object to merge with

        Returns
        -------
        IndexBatch
            Merged index batch
        """
        # Basic cross-checks
        assert (
            self.offsets == index_batch.offsets
        ).all(), "Both index batches should point to the same tensor."

        # Stack the indexes entry-wise in the batch
        indexes, single_counts = [], []
        for b in range(self.batch_size):
            if self.is_list:
                lower, upper = self.edges[b], self.edges[b + 1]
                indexes.extend(self.index_list[lower:upper])
                single_counts.extend(self.single_counts[lower:upper])

                lower, upper = index_batch.edges[b], index_batch.edges[b + 1]
                indexes.extend(index_batch.index_list[lower:upper])
                single_counts.extend(index_batch.single_counts[lower:upper])

            else:
                lower, upper = self.edges[b], self.edges[b + 1]
                indexes.append(self.index[lower:upper])

                lower, upper = index_batch.edges[b], index_batch.edges[b + 1]
                indexes.append(index_batch.index[lower:upper])

        counts = self.counts + index_batch.counts

        if self.is_list:
            return IndexBatch(indexes, self.offsets, counts, single_counts)
        else:
            return IndexBatch(indexes, self.offsets, counts)

    def to_numpy(self):
        """Cast underlying index to a `np.ndarray` and return a new instance.

        Returns
        -------
        TensorBatch
            New `TensorBatch` object with an underlying np.ndarray tensor.
        """
        # If the underlying data is of the right type, nothing to do
        if self.is_numpy:
            return self

        if not self.is_list:
            data = self._to_numpy(self.data)
        else:
            data = np.empty(len(self.data), dtype=object)
            for i, d in enumerate(self.data):
                data[i] = self._to_numpy(d)

        offsets = self._to_numpy(self.offsets)
        counts = self._to_numpy(self.counts)

        single_counts = None
        if self.is_list:
            single_counts = self._to_numpy(self.single_counts)

        return IndexBatch(data, offsets, counts, single_counts)

    def to_tensor(self, dtype=None, device=None):
        """Cast underlying index to a `torch.tensor` and return a new instance.

        Parameters
        ----------
        dtype : torch.dtype, optional
            Data type of the tensor to create
        device : torch.device, optional
            Device on which to put the tensor

        Returns
        -------
        TensorBatch
            New `TensorBatch` object with an underlying np.ndarray tensor.
        """
        # If the underlying data is of the right type, nothing to do
        if not self.is_numpy:
            return self

        if not self.is_list:
            data = self._to_tensor(self.data, dtype, device)
        else:
            data = np.empty(len(data), dtype=object)
            for i, d in enumerate(self.data):
                data[i] = self._to_tensor(d, dtype, device)

        offsets = self._to_tensor(self.offsets, dtype, device)
        counts = self._to_tensor(self.counts, dtype, device)

        single_counts = None
        if self.is_list:
            single_counts = self._to_tensor(self.single_counts, dtype, device)

        return IndexBatch(data, offsets, counts, single_counts)



"""Defines a layer that converts logit output into an evidential output."""

import torch
from torch import nn

from .mlp import MLP


class EvidentialModel(nn.Module):
    """Model which produces evidential predictions with an MLP backbone."""

    def __init__(self, in_channels, mlp, eps=0.0, logspace=False):
        """Initialize the evidential network.

        Parameters
        ----------
        in_channels : int
            Number of features from the upstream feature extractor
        mlp : dict
            MLP configuration dictionary
        eps : float, default 0.0
            Offset to apply to the softplus output
        logspace : bool, default False
            Whether to take the sigmoid of gamma, or not
        """
        # Initialize the parent class
        super().__init__()

        # Initialize the MLP backbone
        self.mlp = MLP(in_channels, **mlp)

        # Initialize the final linear layer
        # TODO: understand why 4, this probably makes sense
        self.linear = nn.Linear(mlp.feature_size, 4)

        # Initialize the output layer parameters
        self.eps = eps
        self.softplus = nn.Softplus
        self.logspace = logspace
        self.gamma = nn.Sigmoid() if logspace else nn.Identity()

    def forward(self, input_feats):
        """Passes a set of features through the evidential model.

        Parameters
        ----------
        input_feats : torch.Tesnor
            (N, F) Tensor of input features

        Results
        -------
        torch.Tensor
            (N, F) Tensor of evidence
        """
        # Pass data through the MLP and the linear layer
        x = self.mlp(input_feats)
        x = self.linear(x)

        # Convert the output to an evidence
        # TODO: Would be nice to understand this
        vab = self.softplus(x[:, :3]) + self.eps
        alpha = torch.clamp(vab[:, 1] + 1.0, min=1.0).view(-1, 1)
        gamma = 2.0 * self.gamma(x[:, 3]).view(-1, 1)
        out = torch.cat(
            [gamma, vab[:, 0].view(-1, 1), alpha, vab[:, 2].view(-1, 1)], dim=1
        )
        if not self.logspace:
            evidence = torch.clamp(out, min=self.eps)
        else:
            evidence = out

        return evidence



"""Manages a singleton instance of a Geometry class."""

import inspect
from typing import Optional

from .base import Geometry
from .factories import geo_factory


class GeoManager:
    """Manages a singleton instance of a Geometry class."""

    _instance: Optional[Geometry] = None

    @classmethod
    def initialize(
        cls, detector: str, tag: Optional[str] = None, version: Optional[str] = None
    ) -> Geometry:
        """Initialize the geometry module for a given detector.

        Parameters
        ----------
        detector : str
            The name of the detector.
        tag : str, optional
            A tag to identify a specific configuration.
        version : str, optional
            A version number for the geometry configuration.

        Returns
        -------
        Geometry
            The initialized geometry instance.
        """
        if cls._instance is not None:
            raise ValueError("Geometry module already initialized.")

        cls._instance = geo_factory(detector, tag, version)

        return cls._instance

    @classmethod
    def initialize_or_get(
        cls, detector: str, tag: Optional[str] = None, version: Optional[str] = None
    ) -> Geometry:
        """Initialize the geometry if needed, or return the existing instance.

        Parameters
        ----------
        detector : str
            The name of the detector.
        tag : str, optional
            A tag to identify a specific configuration.
        version : str, optional
            A version number for the geometry configuration.
        """
        # If the geometry is not initialized, initialize it
        if cls._instance is None:
            cls._instance = geo_factory(detector, tag, version)
            return cls._instance

        # If the geometry is already initialized, check if it matches the
        # requested configuration. If it does not match, initialize it again
        current = cls._instance
        if current.name != detector or current.tag != tag or current.version != version:
            cls._instance = geo_factory(detector, tag, version)

        return cls._instance

    @classmethod
    def is_initialized(cls) -> bool:
        """Check if the geometry instance is initialized.

        Returns
        -------
        bool
            True if the geometry instance is initialized, False otherwise.
        """
        return cls._instance is not None

    @classmethod
    def get_instance(cls) -> Geometry:
        """Get the current geometry instance.

        Parameters
        ----------
        raise_error : bool, optional
            Whether to raise an error if the instance is not initialized.

        Returns
        -------
        Geometry
            The current geometry instance.
        """
        if cls._instance is None:
            # Raise an error with detailed information about the call site
            frame_info = inspect.stack()[1]
            frame = frame_info.frame
            func = frame.f_code.co_name
            cls_obj = frame.f_locals.get("self")
            class_name = cls_obj.__class__.__name__ if cls_obj else None

            location = f"{class_name}.{func}()" if class_name else f"{func}()"

            raise ValueError(
                "Geometry singleton instance is not initialized.\n"
                f"Attempted access from: {location}\n"
                f"File: {frame_info.filename}:{frame_info.lineno}\n\n"
                "If using the Driver, include a `geo` block in the configuration.\n"
                "If running standalone, initialize geometry with either:\n"
                "    GeoManager.initialize(detector='your_detector_name')\n"
                "    GeoManager.initialize_or_get(detector='your_detector_name')\n"
                "before calling geometry-dependent modules."
            )

        return cls._instance

    @classmethod
    def get_instance_if_initialized(cls) -> Optional[Geometry]:
        """Get the current geometry instance if initialized.

        Returns
        -------
        Optional[Geometry]
            The current geometry instance, or None if not initialized.
        """
        return cls._instance

    @classmethod
    def reset(cls) -> None:
        """Reset the geometry instance (useful for testing)."""
        cls._instance = None
